# Tencent is pleased to support the open source community by making Tars available.
#
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#

import struct

from .__util import util
from .exception import *


class BinBuffer:
    def __init__(self, buff=bytes()):
        self.buffer = buff
        self.position = 0

    def writeBuf(self, buff):
        self.buffer += buff

    def getBuffer(self):
        return self.buffer

    def length(self):
        return len(self.buffer)


class DataHead:
    EN_INT8 = 0
    EN_INT16 = 1
    EN_INT32 = 2
    EN_INT64 = 3
    EN_FLOAT = 4
    EN_DOUBLE = 5
    EN_STRING1 = 6
    EN_STRING4 = 7
    EN_MAP = 8
    EN_LIST = 9
    EN_STRUCTBEGIN = 10
    EN_STRUCTEND = 11
    EN_ZERO = 12
    EN_BYTES = 13

    @staticmethod
    def writeTo(buff, tag, vtype):
        if tag < 15:
            helper = (tag << 4) | vtype
            buff.writeBuf(struct.pack('!B', helper))
        else:
            helper = (0xF0 | vtype) << 8 | tag
            buff.writeBuf(struct.pack('!H', helper))


class TarsOutputStream(object):
    def __init__(self):
        self.__buffer = BinBuffer()

    def __writeBoolean(self, tag, value):
        self.__writeInt8(tag, int(value))

    def __writeInt8(self, tag, value):
        if value == 0:
            DataHead.writeTo(self.__buffer, tag, DataHead.EN_ZERO)
        else:
            DataHead.writeTo(self.__buffer, tag, DataHead.EN_INT8)
            self.__buffer.writeBuf(struct.pack('!b', value))

    def __writeInt16(self, tag, value):
        if value >= -128 and value <= 127:
            self.__writeInt8(tag, value)
        else:
            DataHead.writeTo(self.__buffer, tag, DataHead.EN_INT16)
            self.__buffer.writeBuf(struct.pack('!h', value))

    def __writeInt32(self, tag, value):
        if value >= -32768 and value <= 32767:
            self.__writeInt16(tag, value)
        else:
            DataHead.writeTo(self.__buffer, tag, DataHead.EN_INT32)
            self.__buffer.writeBuf(struct.pack('!i', value))

    def __writeInt64(self, tag, value):
        if value >= (-2147483648) and value <= 2147483647:
            self.__writeInt32(tag, value)
        else:
            DataHead.writeTo(self.__buffer, tag, DataHead.EN_INT64)
            self.__buffer.writeBuf(struct.pack('!q', value))

    def __writeFloat(self, tag, value):
        DataHead.writeTo(self.__buffer, tag, DataHead.EN_FLOAT)
        self.__buffer.writeBuf(struct.pack('!f', value))

    def __writeDouble(self, tag, value):
        DataHead.writeTo(self.__buffer, tag, DataHead.EN_DOUBLE)
        self.__buffer.writeBuf(struct.pack('!d', value))

    def __writeString(self, tag, value):
        length = len(value)
        if length <= 255:
            DataHead.writeTo(self.__buffer, tag, DataHead.EN_STRING1)
            self.__buffer.writeBuf(struct.pack('!B', length))
            self.__buffer.writeBuf(str.encode(value))
        else:
            DataHead.writeTo(self.__buffer, tag, DataHead.EN_STRING4)
            self.__buffer.writeBuf(struct.pack('!I', length))
            self.__buffer.writeBuf(str.encode(value))

    def __writeBytes(self, tag, value):
        DataHead.writeTo(self.__buffer, tag, DataHead.EN_BYTES)
        DataHead.writeTo(self.__buffer, 0,   DataHead.EN_INT8)
        length = len(value)
        self.__writeInt32(0, length)
        self.__buffer.buffer += value
        self.__buffer.position += length

    def __writeMap(self, coder, tag, value):
        DataHead.writeTo(self.__buffer, tag, DataHead.EN_MAP)
        self.__writeInt32(0, len(value))
        for key in value:
            self.write(coder.ktype, 0, key)
            self.write(coder.vtype, 1, value.get(key))

    def __writeVector(self, coder, tag, value):
        DataHead.writeTo(self.__buffer, tag, DataHead.EN_LIST)
        n = len(value)
        self.__writeInt32(0, n)
        for i in range(0, n):
            self.write(value.vtype, 0, value[i])

    def __writeStruct(self, coder, tag, value):
        DataHead.writeTo(self.__buffer, tag, DataHead.EN_STRUCTBEGIN)
        value.writeTo(self, value)
        DataHead.writeTo(self.__buffer, 0,   DataHead.EN_STRUCTEND)

    def write(self, coder, tag, value):
        if coder.__tars_index__ == 999:
            self.__writeBoolean(tag, value)
        elif coder.__tars_index__ == 0:
            self.__writeInt8(tag, value)
        elif coder.__tars_index__ == 1:
            self.__writeInt16(tag, value)
        elif coder.__tars_index__ == 2:
            self.__writeInt32(tag, value)
        elif coder.__tars_index__ == 3:
            self.__writeInt64(tag, value)
        elif coder.__tars_index__ == 4:
            self.__writeFloat(tag, value)
        elif coder.__tars_index__ == 5:
            self.__writeDouble(tag, value)
        elif coder.__tars_index__ == 13:
            self.__writeBytes(tag, value)
        elif coder.__tars_index__ == 67:
            self.__writeString(tag, value)
        elif coder.__tars_index__ == 8:
            self.__writeMap(coder, tag, value)
        elif coder.__tars_index__ == 9:
            self.__writeVector(coder, tag, value)
        elif coder.__tars_index__ == 1011:
            self.__writeStruct(coder, tag, value)
        else:
            raise TarsTarsUnsupportType(
                "tars unsupport data type:" % coder.__tars_index__)

    def getBuffer(self):
        return self.__buffer.getBuffer()

    def printHex(self):
        util.printHex(self.__buffer.getBuffer())


class TarsInputStream(object):
    def __init__(self, buff):
        self.__buffer = BinBuffer(buff)

    def __peekFrom(self):
        helper, = struct.unpack_from(
            '!B', self.__buffer.buffer, self.__buffer.position)
        t = (helper & 0xF0) >> 4
        p = (helper & 0x0F)
        l = 1
        if t >= 15:
            l = 2
            t, = struct.unpack_from(
                '!B', self.__buffer.buffer, self.__buffer.position + 1)
        return (t, p, l)

    def __readFrom(self):
        t, p, l = self.__peekFrom()
        self.__buffer.position += l
        return (t, p, l)

    def __skipToStructEnd(self):
        t, p, l = self.__readFrom()
        while p != DataHead.EN_STRUCTEND:
            self.__skipField(p)
            t, p, l = self.__readFrom()

    def __skipField(self, p):
        if p == DataHead.EN_INT8:
            self.__buffer.position += 1
        elif p == DataHead.EN_INT16:
            self.__buffer.position += 2
        elif p == DataHead.EN_INT32:
            self.__buffer.position += 4
        elif p == DataHead.EN_INT64:
            self.__buffer.position += 8
        elif p == DataHead.EN_FLOAT:
            self.__buffer.position += 4
        elif p == DataHead.EN_DOUBLE:
            self.__buffer.position += 8
        elif p == DataHead.EN_STRING1:
            length, = struct.unpack_from(
                '!B', self.__buffer.buffer, self.__buffer.position)
            self.__buffer.position += length + 1
        elif p == DataHead.EN_STRING4:
            length, = struct.unpack_from(
                '!i', self.__buffer.buffer, self.__buffer.position)
            self.__buffer.position += length + 4
        elif p == DataHead.EN_MAP:
            size = self.__readInt32(0, True)
            for i in range(0, size * 2):
                ti, pi, li = self.__readFrom()
                self.__skipField(pi)
        elif p == DataHead.EN_LIST:
            size = self.__readInt32(0, True)
            for i in range(0, size):
                ti, pi, li = self.__readFrom()
                self.__skipField(pi)
        elif p == DataHead.EN_BYTES:
            ti, pi, li = self.__readFrom()
            if pi != DataHead.EN_INT8:
                raise TarsTarsDecodeInvalidValue(
                    "skipField with invalid type, type value: %d, %d." % (p, pi))
            size = self.__readInt32(0, True)
            self.__buffer.position += size
        elif p == DataHead.EN_STRUCTBEGIN:
            self.__skipToStructEnd()
        elif p == DataHead.EN_STRUCTEND:
            pass
            #self.__buffer.position += length + 1;
        elif p == DataHead.EN_ZERO:
            pass
            #self.__buffer.position += length + 1;
        else:
            raise TarsTarsDecodeMismatch(
                "skipField with invalid type, type value:%d" % p)

    def __skipToTag(self, tag):
        length = self.__buffer.length()
        while self.__buffer.position < length:
            t, p, l = self.__peekFrom()
            if tag <= t or p == DataHead.EN_STRUCTEND:
                return False if (p == DataHead.EN_STRUCTEND) else (t == tag)

            self.__buffer.position += l
            self.__skipField(p)
        return False

    def __readBoolean(self, tag, require, default=None):
        v = self.__readInt8(tag, require)
        if v is None:
            return default
        else:
            return (v != 0)

    def __readInt8(self, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p == DataHead.EN_ZERO:
                return 0
            elif p == DataHead.EN_INT8:
                value, = struct.unpack_from(
                    '!b', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 1
                return value
            else:
                raise TarsTarsDecodeMismatch(
                    "read 'Char' type mismatch, tag: %d , get type: %d." % (tag, p))
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def __readInt16(self, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p == DataHead.EN_ZERO:
                return 0
            elif p == DataHead.EN_INT8:
                value, = struct.unpack_from(
                    '!b', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 1
                return value
            elif p == DataHead.EN_INT16:
                value, = struct.unpack_from(
                    '!h', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 2
                return value
            else:
                raise TarsTarsDecodeMismatch(
                    "read 'Short' type mismatch, tag: %d , get type: %d." % (tag, p))
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def __readInt32(self, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p == DataHead.EN_ZERO:
                return 0
            elif p == DataHead.EN_INT8:
                value, = struct.unpack_from(
                    '!b', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 1
                return value
            elif p == DataHead.EN_INT16:
                value, = struct.unpack_from(
                    '!h', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 2
                return value
            elif p == DataHead.EN_INT32:
                value, = struct.unpack_from(
                    '!i', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 4
                return value
            else:
                raise TarsTarsDecodeMismatch(
                    "read 'Int32' type mismatch, tag: %d, get type: %d." % (tag, p))
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def __readInt64(self, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p == DataHead.EN_ZERO:
                return 0
            elif p == DataHead.EN_INT8:
                value, = struct.unpack_from(
                    '!b', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 1
                return value
            elif p == DataHead.EN_INT16:
                value, = struct.unpack_from(
                    '!h', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 2
                return value
            elif p == DataHead.EN_INT32:
                value, = struct.unpack_from(
                    '!i', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 4
                return value
            elif p == DataHead.EN_INT64:
                value, = struct.unpack_from(
                    '!q', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 8
                return value
            else:
                raise TarsTarsDecodeMismatch(
                    "read 'Int64' type mismatch, tag: %d, get type: %d." % (tag, p))
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def __readString(self, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p == DataHead.EN_STRING1:
                length, = struct.unpack_from(
                    '!B', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 1
                value, = struct.unpack_from(
                    str(length) + "s", self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += length
                return value
            elif p == DataHead.EN_STRING4:
                length, = struct.unpack_from(
                    '!i', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 4
                value, = struct.unpack_from(
                    str(length) + "s", self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += length
                return value
            else:
                raise TarsTarsDecodeMismatch(
                    "read 'string' type mismatch, tag: %d, get type: %d." % (tag, p))
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def __readBytes(self, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p == DataHead.EN_BYTES:
                ti, pi, li = self.__readFrom()
                if pi != DataHead.EN_INT8:
                    raise TarsTarsDecodeMismatch(
                        "type mismatch, tag: %d, type: %d, %d" % (tag, p, pi))
                size = self.__readInt32(0, True)
                value, = struct.unpack_from(
                    str(size) + 's', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += size
                return value
            else:
                raise TarsTarsDecodeMismatch(
                    "type mismatch, tag: %d, type: %d" % (tag, p))
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def __readFloat(self, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p == DataHead.EN_ZERO:
                return 0
            elif p == DataHead.EN_FLOAT:
                value, = struct.unpack_from(
                    '!f', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 4
                return value
            else:
                raise TarsTarsDecodeMismatch(
                    "read 'Float' type mismatch, tag: %d, get type: %d." % (tag, p))
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def __readDouble(self, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p == DataHead.EN_ZERO:
                return 0
            elif p == DataHead.EN_FLOAT:
                value, = struct.unpack_from(
                    '!f', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 4
                return value
            elif p == DataHead.EN_DOUBLE:
                value, = struct.unpack_from(
                    '!d', self.__buffer.buffer, self.__buffer.position)
                self.__buffer.position += 8
                return value
            else:
                raise TarsTarsDecodeMismatch(
                    "read 'Double' type mismatch, tag: %d, get type: %d." % (tag, p))
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def __readStruct(self, coder, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p != DataHead.EN_STRUCTBEGIN:
                raise TarsTarsDecodeMismatch(
                    "read 'struct' type mismatch, tag: %d, get type: %d." % (tag, p))
            value = coder.readFrom(self)
            self.__skipToStructEnd()
            return value
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def __readMap(self, coder, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p == DataHead.EN_MAP:
                size = self.__readInt32(0, True)
                omap = coder()
                for i in range(0, size):
                    k = self.read(coder.ktype, 0, True)
                    v = self.read(coder.vtype, 1, True)
                    omap[k] = v
                return omap
            else:
                raise TarsTarsDecodeMismatch(
                    "read 'map' type mismatch, tag: %d, get type: %d." % (tag, p))
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def __readVector(self, coder, tag, require, default=None):
        if self.__skipToTag(tag):
            t, p, l = self.__readFrom()
            if p == DataHead.EN_LIST:
                size = self.__readInt32(0, True)
                value = coder()
                for i in range(0, size):
                    k = self.read(coder.vtype, 0, True)
                    value.append(k)
                return value
            else:
                raise TarsTarsDecodeMismatch(
                    "read 'vector' type mismatch, tag: %d, get type: %d." % (tag, p))
        elif require:
            raise TarsTarsDecodeRequireNotExist(
                "require field not exist, tag: %d" % tag)
        return default

    def read(self, coder, tag, require, default=None):
        if coder.__tars_index__ == 999:
            return self.__readBoolean(tag, require, default)
        elif coder.__tars_index__ == 0:
            return self.__readInt8(tag, require, default)
        elif coder.__tars_index__ == 1:
            return self.__readInt16(tag, require, default)
        elif coder.__tars_index__ == 2:
            return self.__readInt32(tag, require, default)
        elif coder.__tars_index__ == 3:
            return self.__readInt64(tag, require, default)
        elif coder.__tars_index__ == 4:
            return self.__readFloat(tag, require, default)
        elif coder.__tars_index__ == 5:
            return self.__readDouble(tag, require, default)
        elif coder.__tars_index__ == 13:
            return self.__readBytes(tag, require, default)
        elif coder.__tars_index__ == 67:
            return self.__readString(tag, require, default)
        elif coder.__tars_index__ == 8:
            return self.__readMap(coder, tag, require, default)
        elif coder.__tars_index__ == 9:
            return self.__readVector(coder, tag, require, default)
        elif coder.__tars_index__ == 1011:
            return self.__readStruct(coder, tag, require, default)
        else:
            raise TarsTarsUnsupportType(
                "tars unsupport data type:" % coder.__tars_index__)

    def printHex(self):
        util.printHex(self.__buffer.buffer)



"""
HTTP Basic Auth python lib from https://github.com/bugov/http-basic-auth
"""

import base64
from typing import Tuple

__version__ = '1.2.0'


class BasicAuthException(Exception):
    """General exception for all http-basic-auth problems
    """

def parse_token(token: str, coding='utf-8') -> Tuple[str, str]:
    """Get login + password tuple from Basic Auth token.
    """
    try:
        b_token = bytes(token, encoding=coding)
    except UnicodeEncodeError as e:
        raise BasicAuthException from e
    except TypeError as e:
        raise BasicAuthException from e

    try:
        auth_pair = base64.b64decode(b_token, validate=True)
    except base64.binascii.Error as e:
        raise BasicAuthException from e

    try:
        (login, password) = auth_pair.split(b':', maxsplit=1)
    except ValueError as e:
        raise BasicAuthException from e

    try:
        return str(login, encoding=coding), str(password, encoding=coding)
    except UnicodeDecodeError as e:
        raise BasicAuthException from e


def generate_token(login: str, password: str, coding='utf-8') -> str:
    """Generate Basic Auth token from login and password
    """
    try:
        b_login = bytes(login, encoding=coding)
        b_password = bytes(password, encoding=coding)
    except UnicodeEncodeError as e:
        raise BasicAuthException from e
    except TypeError as e:
        raise BasicAuthException from e

    if b':' in b_login:
        raise BasicAuthException

    b_token = base64.b64encode(b'%b:%b' % (b_login, b_password))

    return str(b_token, encoding=coding)


def parse_header(header_value: str, coding='utf-8') -> Tuple[str, str]:
    """Get login + password tuple from Basic Auth header value.
    """
    if header_value is None:
        raise BasicAuthException

    try:
        basic_prefix, token = header_value.strip().split(maxsplit=1)
    except AttributeError as e:
        raise BasicAuthException from e
    except ValueError as e:
        raise BasicAuthException from e

    if basic_prefix.lower() != 'basic':
        raise BasicAuthException

    return parse_token(token, coding=coding)


def generate_header(login: str, password: str, coding='utf-8') -> str:
    """Generate Basic Auth header value from login and password
    """
    return 'Basic %s' % generate_token(login, password, coding=coding)


import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor

from . import plugins
from biliup.config import config
from biliup.engine import Plugin, invert_dict
from biliup.engine.event import EventManager, Event
from .common.timer import Timer
from .common.tools import NamedLock

logger = logging.getLogger('biliup')


def create_event_manager():
    pool1_size = config.get('pool1_size', 5)
    pool2_size = config.get('pool2_size', 3)
    pool = {
        'Asynchronous1': ThreadPoolExecutor(pool1_size, thread_name_prefix='Asynchronous1'),
        'Asynchronous2': ThreadPoolExecutor(pool2_size, thread_name_prefix='Asynchronous2'),
        # 'Asynchronous3': ThreadPoolExecutor(2, thread_name_prefix='Asynchronous3'),
    }
    # 初始化事件管理器
    app = EventManager(config, pool)
    app.context['url_upload_count'] = {}
    # 正在上传的文件 用于同时上传一个url的时候过滤掉正在上传的
    app.context['upload_filename'] = []
    return app


event_manager = create_event_manager()
context = event_manager.context


async def singleton_check(platform, name, url):
    from biliup.handler import PRE_DOWNLOAD, UPLOAD
    context['url_upload_count'].setdefault(url, 0)
    if context['PluginInfo'].url_status[url] == 1:
        logger.debug(f'{url} 正在下载中，跳过检测')
        return

    event_manager.send_event(Event(UPLOAD, ({'name': name, 'url': url},)))
    if await platform(name, url).acheck_stream(True):
        # 需要等待上传文件列表检索完成后才可以开始下次下载
        with NamedLock(f'upload_file_list_{name}'):
            event_manager.send_event(Event(PRE_DOWNLOAD, args=(name, url,)))


async def shot(event):
    index = 0
    while True:
        if not len(event.url_list):
            logger.info(f"{event}没有任务，退出")
            return
        if index >= len(event.url_list):
            index = 0
            continue
        cur = event.url_list[index]
        try:
            await singleton_check(event, context['PluginInfo'].inverted_index[cur], cur)
            index += 1
            skip = context['PluginInfo'].url_status[cur] == 1 and index < len(event.url_list)
            if skip:  # 在一次 url_list 内，如果 url 正在下载，则跳过本次等待以加快下一个检测
                continue
        except Exception:
            logger.exception('shot')
        await asyncio.sleep(config.get('event_loop_interval', 30))


@event_manager.server()
class PluginInfo:
    def __init__(self, streamers):
        streamer_url = {k: v['url'] for k, v in streamers.items()}
        self.inverted_index = invert_dict(streamer_url)
        urls = list(self.inverted_index.keys())
        self.checker = Plugin(plugins).sorted_checker(urls)
        self.url_status = dict.fromkeys(self.inverted_index, 0)
        self.coroutines = dict.fromkeys(self.checker)
        self.init_tasks()

    def add(self, name, url):
        temp = Plugin(plugins).inspect_checker(url)
        key = temp.__name__
        if key in self.checker:
            self.checker[key].url_list.append(url)
        else:
            temp.url_list = [url]
            self.checker[key] = temp
            from .engine.download import BatchCheck
            if issubclass(temp, BatchCheck):
                # 如果支持批量检测
                self.batch_check_task(temp)
            else:
                self.coroutines[key] = asyncio.create_task(shot(temp))
        self.inverted_index[url] = name
        self.url_status[url] = 0

    def delete(self, url):
        if not url in self.inverted_index:
            return
        del self.inverted_index[url]
        exec_del = False
        for key, value in self.checker.items():
            if url in value.url_list:
                if len(value.url_list) == 1:
                    exec_del = key
                else:
                    value.url_list.remove(url)
        if exec_del:
            del self.checker[exec_del]
            self.coroutines[exec_del].cancel()
            del self.coroutines[exec_del]

    def init_tasks(self):
        from .engine.download import BatchCheck

        for key, plugin in self.checker.items():
            if issubclass(plugin, BatchCheck):
                # 如果支持批量检测
                self.batch_check_task(plugin)
                continue
            self.coroutines[key] = asyncio.create_task(shot(plugin))

    def batch_check_task(self, plugin):
        from biliup.handler import PRE_DOWNLOAD

        async def check_timer():
            name = None
            # 如果支持批量检测
            try:
                async for turl in plugin.abatch_check(plugin.url_list):
                    context['url_upload_count'].setdefault(turl, 0)
                    for k, v in config['streamers'].items():
                        if v.get("url", "") == turl:
                            name = k
                    event_manager.send_event(Event(PRE_DOWNLOAD, args=(name, turl,)))
            except Exception:
                logger.exception('batch_check_task')

        timer = Timer(func=check_timer, interval=30)
        self.coroutines[plugin.__name__] = asyncio.create_task(timer.astart())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# filename: __trans.py

# Tencent is pleased to support the open source community by making Tars available.
#
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#


'''
@version: 0.01
@brief: 网络相关模块
'''

import errno
import select
import socket
import threading

from .__TimeoutQueue import ReqMessage
from .__logger import tarsLogger


class EndPointInfo:
    '''
    @brief: 保存每个连接端口的信息
    '''
    SOCK_TCP = 'TCP'
    SOCK_UDP = 'UDP'

    def __init__(self,
                 ip='',
                 port=0,
                 timeout=-1,
                 weight=0,
                 weightType=0,
                 connType=SOCK_TCP):
        self.__ip = ip
        self.__port = port
        self.__timeout = timeout
        self.__connType = connType
        self.__weightType = weightType
        self.__weight = weight

    def getIp(self):
        return self.__ip

    def getPort(self):
        return self.__port

    def getConnType(self):
        '''
        @return: 传输层连接类型
        @rtype: EndPointInfo.SOCK_TCP 或 EndPointInfo.SOCK_UDP
        '''
        return self.__connType

    def getWeightType(self):
        return self.__weightType

    def getWeight(self):
        return self.__weight

    def __str__(self):
        return '%s %s:%s %d:%d' % (self.__connType, self.__ip, self.__port, self.__weightType, self.__weight)


class Transceiver:
    '''
    @brief: 网络传输基类，提供网络send/recv接口
    '''
    CONNECTED = 0
    CONNECTING = 1
    UNCONNECTED = 2

    def __init__(self, endPointInfo):
        tarsLogger.debug('Transceiver:__init__, %s', endPointInfo)
        self.__epi = endPointInfo
        self.__sock = None
        self.__connStatus = Transceiver.UNCONNECTED
        self.__connFailed = False
        # 这两个变量要给子类用，不能用name mangling隐藏
        self._sendBuff = ''
        self._recvBuf = ''

    def __del__(self):
        tarsLogger.debug('Transceiver:__del__')
        self.close()

    def getSock(self):
        '''
        @return: socket对象
        @rtype: socket.socket
        '''
        return self.__sock

    def getFd(self):
        '''
        @brief: 获取socket的文件描述符
        @return: 如果self.__sock没有建立返回-1
        @rtype: int
        '''
        if self.__sock:
            return self.__sock.fileno()
        else:
            return -1

    def getEndPointInfo(self):
        '''
        @return: 端口信息
        @rtype: EndPointInfo
        '''
        return self.__epi

    def isValid(self):
        '''
        @return: 是否创建了socket
        @rtype: bool
        '''
        return self.__sock is not None

    def hasConnected(self):
        '''
        @return: 是否连接上了
        @rtype: bool
        '''
        return self.isValid() and self.__connStatus == Transceiver.CONNECTED

    def isConnFailed(self):
        '''
        @return: 是否连接失败
        @rtype: bool
        '''
        return self.__connFailed

    def isConnecting(self):
        '''
        @return: 是否正在连接
        @rtype: bool
        '''
        return self.isValid() and self.__connStatus == Transceiver.CONNECTING

    def setConnFailed(self):
        '''
        @brief: 设置为连接失败
        @return: None
        @rtype: None
        '''
        self.__connFailed = True
        self.__connStatus = Transceiver.UNCONNECTED

    def setConnected(self):
        '''
        @brief: 设置为连接完
        @return: None
        @rtype: None
        '''
        self.__connFailed = False
        self.__connStatus = Transceiver.CONNECTED

    def close(self):
        '''
        @brief: 关闭连接
        @return: None
        @rtype: None
        @note: 多次调用不会有问题
        '''
        tarsLogger.debug('Transceiver:close')
        if not self.isValid():
            return
        self.__sock.close()
        self.__sock = None
        self.__connStatus = Transceiver.UNCONNECTED
        self.__connFailed = False
        self._sendBuff = ''
        self._recvBuf = ''
        tarsLogger.info('trans close : %s' % self.__epi)

    def writeToSendBuf(self, msg):
        '''
        @brief: 把数据添加到send buffer里
        @param msg: 发送的数据
        @type msg: str
        @return: None
        @rtype: None
        @note: 没有加锁，多线程调用会有race conditions
        '''
        self._sendBuff += msg

    def recv(self, bufsize, flag=0):
        raise NotImplementedError()

    def send(self, buf, flag=0):
        raise NotImplementedError()

    def doResponse(self):
        raise NotImplementedError()

    def doRequest(self):
        '''
        @brief: 将请求数据发送出去
        @return: 发送的字节数
        @rtype: int
        '''
        tarsLogger.debug('Transceiver:doRequest')
        if not self.isValid():
            return -1

        nbytes = 0
        buf = buffer(self._sendBuff)
        while True:
            if not buf:
                break
            ret = self.send(buf[nbytes:])
            if ret > 0:
                nbytes += ret
            else:
                break

        # 发送前面的字节后将后面的字节拷贝上来
        self._sendBuff = buf[nbytes:]
        return nbytes

    def reInit(self):
        '''
        @brief: 初始化socket，并连接服务器
        @return: 成功返回0，失败返回-1
        @rtype: int
        '''
        tarsLogger.debug('Transceiver:reInit')
        assert(self.isValid() is False)
        if self.__epi.getConnType() != EndPointInfo.SOCK_TCP:
            return -1
        try:
            self.__sock = socket.socket()
            self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.__sock.setblocking(0)
            self.__sock.connect((self.__epi.getIp(), self.__epi.getPort()))
            self.__connStatus = Transceiver.CONNECTED
        except socket.error as msg:
            if msg.errno == errno.EINPROGRESS:
                self.__connStatus = Transceiver.CONNECTING
            else:
                tarsLogger.info('reInit, %s, faild!, %s',
                                self.__epi, msg)
                self.__sock = None
                return -1
        tarsLogger.info('reInit, connect: %s, fd: %d',
                        self.__epi, self.getFd())
        return 0


class TcpTransceiver(Transceiver):
    '''
    @brief: TCP传输实现
    '''

    def send(self, buf, flag=0):
        '''
        @brief: 实现tcp的发送
        @param buf: 发送的数据
        @type buf: str
        @param flag: 发送标志
        @param flag: int
        @return: 发送字节数
        @rtype: int
        '''
        tarsLogger.debug('TcpTransceiver:send')
        if not self.isValid():
            return -1

        nbytes = 0
        try:
            nbytes = self.getSock().send(buf, flag)
            tarsLogger.info('tcp send, fd: %d, %s, len: %d',
                            self.getFd(), self.getEndPointInfo(), nbytes)
        except socket.error as msg:
            if msg.errno != errno.EAGAIN:
                tarsLogger.error('tcp send, fd: %d, %s, fail!, %s, close',
                                 self.getFd(), self.getEndPointInfo(), msg)
                self.close()
                return 0
        return nbytes

    def recv(self, bufsize, flag=0):
        '''
        @brief: 实现tcp的recv
        @param bufsize: 接收大小
        @type bufsize: int
        @param flag: 接收标志
        @param flag: int
        @return: 接收的内容，接收出错返回None
        @rtype: str
        '''
        tarsLogger.debug('TcpTransceiver:recv')
        assert(self.isValid())

        buf = ''
        try:
            buf = self.getSock().recv(bufsize, flag)
            if len(buf) == 0:
                tarsLogger.info('tcp recv, fd: %d, %s, recv 0 bytes, close',
                                self.getFd(), self.getEndPointInfo())
                self.close()
                return None
        except socket.error as msg:
            if msg.errno != errno.EAGAIN:
                tarsLogger.info('tcp recv, fd: %d, %s, faild!, %s, close',
                                self.getFd(), self.getEndPointInfo(), msg)
                self.close()
                return None

        tarsLogger.info('tcp recv, fd: %d, %s, nbytes: %d',
                        self.getFd(), self.getEndPointInfo(), len(buf))
        return buf

    def doResponse(self):
        '''
        @brief: 处理接收的数据
        @return: 返回响应报文的列表，如果出错返回None
        @rtype: list: ResponsePacket
        '''
        tarsLogger.debug('TcpTransceiver:doResponse')
        if not self.isValid():
            return None

        bufs = [self._recvBuf]
        while True:
            buf = self.recv(8292)
            if not buf:
                break
            bufs.append(buf)
        self._recvBuf = ''.join(bufs)
        tarsLogger.info('tcp doResponse, fd: %d, recvbuf: %d',
                        self.getFd(), len(self._recvBuf))

        if not self._recvBuf:
            return None

        rsplist = None
        try:
            rsplist, bufsize = ReqMessage.unpackRspList(self._recvBuf)
            self._recvBuf = self._recvBuf[bufsize:]
        except Exception as msg:
            tarsLogger.error(
                'tcp doResponse, fd: %d, %s, tcp recv unpack error: %s',
                self.getFd(), self.getEndPointInfo(), msg)
            self.close()

        return rsplist


class FDReactor(threading.Thread):
    '''
    @brief: 监听FD事件并解发注册的handle
    '''

    def __init__(self):
        tarsLogger.debug('FDReactor:__init__')
        # threading.Thread.__init__(self)
        super(FDReactor, self).__init__()
        self.__terminate = False
        self.__ep = None
        self.__shutdown = None
        # {fd : adapterproxy}
        self.__adapterTab = {}

    def __del__(self):
        tarsLogger.debug('FDReactor:__del__')
        self.__ep.close()
        self.__shutdown.close()
        self.__ep = None
        self.__shutdown = None

    def initialize(self):
        '''
        @brief: 初始化，使用FDReactor前必须调用
        @return: None
        @rtype: None
        '''
        tarsLogger.debug('FDReactor:initialize')
        self.__ep = select.epoll()
        self.__shutdown = socket.socket()
        self.__ep.register(self.__shutdown.fileno(),
                           select.EPOLLET | select.EPOLLIN)
        tarsLogger.debug('FDReactor init, shutdown fd : %d',
                         self.__shutdown.fileno())

    def terminate(self):
        '''
        @brief: 结束FDReactor的线程
        @return: None
        @rtype: None
        '''
        tarsLogger.debug('FDReactor:terminate')
        self.__terminate = True
        self.__ep.modify(self.__shutdown.fileno(), select.EPOLLOUT)
        self.__adapterTab = {}

    def handle(self, adapter, events):
        '''
        @brief: 处理epoll事件
        @param adapter: 事件对应的adapter
        @type adapter: AdapterProxy
        @param events: epoll事件
        @param events: int
        @return: None
        @rtype: None
        '''
        tarsLogger.debug('FDReactor:handle events : %d', events)
        assert(adapter)

        try:
            if events == 0:
                return

            if events & (select.EPOLLERR | select.EPOLLHUP):
                tarsLogger.debug('FDReactor::handle EPOLLERR or EPOLLHUP: %s',
                                 adapter.trans().getEndPointInfo())
                adapter.trans().close()
                return

            if adapter.shouldCloseTrans():
                tarsLogger.debug('FDReactor::handle should close trans: %s',
                                 adapter.trans().getEndPointInfo())
                adapter.setCloseTrans(False)
                adapter.trans().close()
                return

            if adapter.trans().isConnecting():
                if not adapter.finishConnect():
                    return

            if events & select.EPOLLIN:
                self.handleInput(adapter)

            if events & select.EPOLLOUT:
                self.handleOutput(adapter)

        except Exception as msg:
            tarsLogger.error('FDReactor handle exception: %s', msg)

    def handleExcept(self):
        pass

    def handleInput(self, adapter):
        '''
        @brief: 处理接收事件
        @param adapter: 事件对应的adapter
        @type adapter: AdapterProxy
        @return: None
        @rtype: None
        '''

        tarsLogger.debug('FDReactor:handleInput')
        if not adapter.trans().isValid():
            return

        rsplist = adapter.trans().doResponse()
        if not rsplist:
            return
        for rsp in rsplist:
            adapter.finished(rsp)

    def handleOutput(self, adapter):
        '''
        @brief: 处理发送事件
        @param adapter: 事件对应的adapter
        @type adapter: AdapterProxy
        @return: None
        @rtype: None
        '''
        tarsLogger.debug('FDReactor:handleOutput')
        if not adapter.trans().isValid():
            return
        while adapter.trans().doRequest() >= 0 and adapter.sendRequest():
            pass

    def notify(self, adapter):
        '''
        @brief: 更新adapter对应的fd的epoll状态
        @return: None
        @rtype: None
        @note: FDReactor使用的epoll是EPOLLET模式，同一事件只通知一次
               希望某一事件再次通知需调用此函数
        '''
        tarsLogger.debug('FDReactor:notify')
        fd = adapter.trans().getFd()
        if fd != -1:
            self.__ep.modify(fd,
                             select.EPOLLET | select.EPOLLOUT | select.EPOLLIN)

    def registerAdapter(self, adapter, events):
        '''
        @brief: 注册adapter
        @param adapter: 收发事件处理类
        @type adapter: AdapterProxy
        @param events: 注册事件
        @type events: int
        @return: None
        @rtype: None
        '''
        tarsLogger.debug('FDReactor:registerAdapter events : %d', events)
        events |= select.EPOLLET
        try:
            self.__ep.unregister(adapter.trans().getFd())
        except:
            pass
        self.__ep.register(adapter.trans().getFd(), events)
        self.__adapterTab[adapter.trans().getFd()] = adapter

    def unregisterAdapter(self, adapter):
        '''
        @brief: 注销adapter
        @param adapter: 收发事件处理类
        @type adapter: AdapterProxy
        @return: None
        @rtype: None
        '''
        tarsLogger.debug('FDReactor:registerAdapter')
        self.__ep.unregister(adapter.trans().getFd())
        self.__adapterTab.pop(adapter.trans().getFd(), None)

    def run(self):
        '''
        @brief: 线程启动函数，循环监听网络事件
        '''
        tarsLogger.debug('FDReactor:run')

        while not self.__terminate:
            try:
                eplist = self.__ep.poll(1)
                if eplist:
                    tarsLogger.debug('FDReactor run get eplist : %s, terminate : %s', str(
                        eplist), self.__terminate)
                if self.__terminate:
                    tarsLogger.debug('FDReactor terminate')
                    break
                for fd, events in eplist:
                    adapter = self.__adapterTab.get(fd, None)
                    if not adapter:
                        continue
                    self.handle(adapter, events)
            except Exception as msg:
                tarsLogger.error('FDReactor run exception: %s', msg)

        tarsLogger.debug('FDReactor:run finished')


if __name__ == '__main__':
    print('hello world')
    epi = EndPointInfo('127.0.0.1', 1313)
    print(epi)
    trans = TcpTransceiver(epi)
    print(trans.getSock())
    print(trans.getFd())
    print(trans.reInit())
    print(trans.isConnecting())
    print(trans.hasConnected())
    buf = 'hello world'
    print(trans.send(buf))
    buf = trans.recv(1024)
    print(buf)
    trans.close()



import logging
import re

logger = logging.getLogger('biliup')


def match1(text, *patterns):
    if len(patterns) == 1:
        pattern = patterns[0]
        match = re.search(pattern, text)
        if match:
            return match.group(1)
        else:
            return None
    else:
        ret = []
        for pattern in patterns:
            match = re.search(pattern, text)
            if match:
                ret.append(match.group(1))
        return ret


def random_user_agent(device: str = 'desktop') -> str:
    import random
    chrome_version = random.randint(100, 120)
    if device == 'mobile':
        android_version = random.randint(9, 14)
        mobile = random.choice([
            'SM-G981B', 'SM-G9910', 'SM-S9080', 'SM-S9110', 'SM-S921B',
            'Pixel 5', 'Pixel 6', 'Pixel 7', 'Pixel 7 Pro', 'Pixel 8',
        ])
        return f'Mozilla/5.0 (Linux; Android {android_version}; {mobile}) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{chrome_version}.0.0.0 Mobile Safari/537.36'
    return f'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{chrome_version}.0.0.0 Safari/537.36'



import base64
import hashlib
import json
import random
import time
from urllib.parse import parse_qs, unquote
from functools import lru_cache

from biliup.common.util import client
from biliup.config import config
from biliup.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from ..plugins import logger, random_user_agent


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www|m)\.)?huya\.com')
class Huya(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.fake_headers['referer'] = url
        self.fake_headers['cookie'] = config.get('user', {}).get('huya_cookie', '')
        self._room_id = url.split('huya.com/')[1].split('?')[0]
        self.huya_danmaku = config.get('huya_danmaku', False)


    async def acheck_stream(self, is_check=False):
        try:
            if self.fake_headers.get('cookie'):
                await self.verify_cookie()
            if not self._room_id.isdigit():
                self._room_id = _get_real_rid(self.url)
            room_profile = await self.get_room_profile(use_api=True)
        except Exception as e:
            logger.error(f"{self.plugin_msg}: {e}")
            return False

        if room_profile['realLiveStatus'] != 'ON' or room_profile['liveStatus'] != 'ON':
            '''
            ON: 直播
            REPLAY: 重播
            OFF: 未开播
            '''
            logger.debug(f"{self.plugin_msg} : 未开播")
            self.raw_stream_url = None
            return False

        if not room_profile['liveData'].get('bitRateInfo'):
            # 主播未推流
            logger.debug(f"{self.plugin_msg} : 未推流")
            return False

        if is_check:
            return True

        huya_max_ratio = config.get('huya_max_ratio', 0)
        if huya_max_ratio:
            try:
                # 最大码率(不含hdr)
                # max_ratio = html_info['data'][0]['gameLiveInfo']['bitRate']
                max_ratio = room_profile['liveData']['bitRate']
                # 可选择的码率
                live_rate_info = json.loads(room_profile['liveData']['bitRateInfo'])
                # 码率信息
                ratio_items = [r.get('iBitRate', 0) if r.get('iBitRate', 0) != 0 else max_ratio for r in live_rate_info]
                # 符合条件的码率
                ratio_in_items = [x for x in ratio_items if x <= huya_max_ratio]
                # 录制码率
                if ratio_in_items:
                    record_ratio = max(ratio_in_items)
                else:
                    record_ratio = max_ratio
            except Exception as e:
                logger.error(f"{self.plugin_msg}: 在确定码率时发生错误 {e}")
                return False

        huya_cdn = config.get('huyacdn', 'AL') # 将于 0.5.0 删除
        perf_cdn = config.get('huya_cdn', huya_cdn).upper() # 0.5.0 允许为空字符串以使用 Api 内的 CDN 优先级
        protocol = 'Hls' if config.get('huya_protocol') == 'Hls' else 'Flv'
        allow_imgplus = config.get('huya_imgplus', True)
        cdn_fallback = config.get('huya_cdn_fallback', False)
        use_api = config.get('huya_mobile_api', False)

        try:
            stream_urls = await self.get_stream_urls(protocol, use_api, allow_imgplus)
        except:
            logger.exception(f"{self.plugin_msg}: 没有可用的链接")
            return False

        cdn_name_list = list(stream_urls.keys())
        if not perf_cdn or perf_cdn not in cdn_name_list:
            logger.warning(f"{self.plugin_msg}: 使用 {cdn_name_list[0]}")
            perf_cdn = cdn_name_list[0]

        # 虎牙直播流只允许连接一次
        if cdn_fallback:
            _url = await self.acheck_url_healthy(stream_urls[perf_cdn])
            if _url is None:
                logger.info(f"{self.plugin_msg}: 提供如下CDN {cdn_name_list}")
                for cdn in cdn_name_list:
                    if cdn == perf_cdn:
                        continue
                    logger.info(f"{self.plugin_msg}: cdn_fallback 尝试 {cdn}")
                    if (await self.acheck_url_healthy(stream_urls[cdn])) is None:
                        continue
                    perf_cdn = cdn
                    logger.info(f"{self.plugin_msg}: CDN 切换为 {perf_cdn}")
                    break
                else:
                    logger.error(f"{self.plugin_msg}: cdn_fallback 所有链接无法使用")
                    return False
            stream_urls = await self.get_stream_urls(protocol, use_api, allow_imgplus)

        # self.room_title = html_info['data'][0]['gameLiveInfo']['introduction']
        self.room_title = room_profile['liveData']['introduction']
        self.raw_stream_url = stream_urls[perf_cdn]

        if huya_max_ratio and record_ratio != max_ratio:
            self.raw_stream_url += f"&ratio={record_ratio}"
        return True


    def danmaku_init(self):
        if self.huya_danmaku:
            self.danmaku = DanmakuClient(self.url, self.gen_download_filename())


    async def get_room_profile(self, use_api=False) -> dict:
        if use_api:
            resp = (await client.get(f"https://mp.huya.com/cache.php?m=Live&do=profileRoom&roomid={self._room_id}", \
                                        headers=self.fake_headers)).json()
            if resp['status'] != 200:
                raise Exception(f"{resp['message']}")
            return resp['data']
        else:
            html = (await client.get(f"https://www.huya.com/{self._room_id}", headers=self.fake_headers)).text
            if '找不到这个主播' in html:
                raise Exception(f"找不到这个主播")
            return json.loads(html.split('stream: ')[1].split('};')[0])


    async def get_stream_urls(self, protocol, use_api=False, allow_imgplus=True) -> dict:
        '''
        返回指定协议的所有CDN流
        '''
        streams = {}
        room_profile = await self.get_room_profile(use_api=use_api)
        if not use_api:
            try:
                stream_info = room_profile['data'][0]['gameStreamInfoList']
            except KeyError:
                raise Exception(f"{room_profile}")
        else:
            stream_info = room_profile['stream']['baseSteamInfoList']
            streams = _dict_sorting(json.loads(room_profile['liveData'].get('mStreamRatioWeb', '{}')))
        stream = stream_info[0]
        stream_name = stream['sStreamName']
        suffix, anti_code = stream[f's{protocol}UrlSuffix'], stream[f's{protocol}AntiCode']
        if not allow_imgplus:
            stream_name = stream_name.replace('-imgplus', '')
        anti_code = self.__build_query(stream_name, anti_code, self.fake_headers['cookie'])
        for stream in stream_info:
            stream_url = f"{stream[f's{protocol}Url']}/{stream_name}.{suffix}?{anti_code}"
            if stream['sCdnType'] in ['HY', 'HUYA', 'HYZJ']: continue
            streams[stream['sCdnType']] = stream_url
        return streams

    @staticmethod
    def __build_query(stream_name, anti_code, cookies=None) -> str:
        url_query = parse_qs(anti_code)
        platform_id = 100
        uid = _get_uid(cookies, stream_name)
        convert_uid = (uid << 8 | uid >> (32 - 8)) & 0xFFFFFFFF
        ws_time = url_query['wsTime'][0]
        ct = int((int(ws_time, 16) + random.random()) * 1000)
        seq_id = uid + int(time.time() * 1000)
        ws_secret_prefix = base64.b64decode(unquote(url_query['fm'][0]).encode()).decode().split('_')[0]
        ws_secret_hash = hashlib.md5(f"{seq_id}|{url_query['ctype'][0]}|{platform_id}".encode()).hexdigest()
        ws_secret = hashlib.md5(f'{ws_secret_prefix}_{convert_uid}_{stream_name}_{ws_secret_hash}_{ws_time}'.encode()).hexdigest()
        # &codec=av1
        # &codec=264
        # &codec=265
        # dMod: wcs-25 浏览器解码信息
        # sdkPcdn: 1_1 第一个1连接次数 第二个1是因为什么连接
        # t: 100 平台信息 100 web
        # sv: 2401090219 版本
        # sdk_sid:  _sessionId sdkInRoomTs 当前毫秒时间

        # return f"wsSecret={ws_secret}&wsTime={ws_time}&seqid={seq_id}&ctype={url_query['ctype'][0]}&ver=1&fs={url_query['fs'][0]}&u={convert_uid}&t={platform_id}&sv=2401090219&sdk_sid={int(time.time() * 1000)}&codec=264"

        anti_code = {
            "wsSecret": ws_secret,
            "wsTime": ws_time,
            "seqid": str(seq_id),
            "ctype": url_query['ctype'][0],
            "fs": url_query['fs'][0],
            "u": convert_uid,
            "t": platform_id,
            "ver": "1",
            "uuid": str(int((ct % 1e10 + random.random()) * 1e3 % 0xffffffff)),
            "sdk_sid": str(int(time.time() * 1000)),
            "codec": "264",
        }
        return '&'.join([f"{k}={v}" for k, v in anti_code.items()])


    async def verify_cookie(self):
        if self.fake_headers['cookie']:
            resp = (await client.post('https://udblgn.huya.com/web/cookie/verify', \
                                    headers=self.fake_headers, data={'appId': 5002})).json()
            if resp.json()['returnCode'] != 0:
                logger.error(f"{self.plugin_msg}: {resp.json()['message']}")
                self.fake_headers['cookie'] = ''

@lru_cache(maxsize=None)
def _get_real_rid(url):
    import requests
    headers = {
        'user-agent': random_user_agent(),
    }
    html = requests.get(url, headers=headers).text
    if '找不到这个主播' in html:
        raise Exception(f"找不到这个主播")
    html_obj = json.loads(html.split('stream: ')[1].split('};')[0])
    return html_obj['data'][0]['gameLiveInfo']['profileRoom']


def _dict_sorting(data: dict) -> dict:
    if data:
        data = {k: v for k, v in data.items() if k not in ['HY', 'HUYA', 'HYZJ']}
        return dict(sorted(data.items(), key=lambda x: x[1], reverse=True))
    return {}


def _get_uid(cookie: str, stream_name: str) -> int:
    if cookie:
        cookie_dict = {k.strip(): v for k, v in (item.split('=') for item in cookie.split(';'))}
        for key in ['udb_uid', 'yyuid']:
            if key in cookie_dict:
                return int(cookie_dict[key])
    if stream_name:
        return int(stream_name.split('-')[0])
    return random.randint(1400000000000, 1499999999999)



import time
from typing import Optional, Dict

import requests

import biliup.common.util
from biliup.config import config
from ..common import tools
from ..common.tools import NamedLock
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from ..plugins import match1, logger

# VALID_URL_BASE = r"https?://(.*?)\.afreecatv\.com/(?P<username>\w+)(?:/\d+)?"
VALID_URL_BASE = r"https?://play\.afreecatv\.com/(?P<username>\w+)(?:/\d+)?"
CHANNEL_API_URL = "https://live.afreecatv.com/afreeca/player_live_api.php"

QUALITIES = ["original", "hd4k", "hd", "sd"]


@Plugin.download(regexp=r"https?://(.*?)\.afreecatv\.com/(?P<username>\w+)(?:/\d+)?")
class AfreecaTV(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        if AfreecaTVUtils.get_cookie():
            self.fake_headers['cookie'] = ';'.join(
                [f"{name}={value}" for name, value in AfreecaTVUtils.get_cookie().items()])

    async def acheck_stream(self, is_check=False):
        try:
            username = match1(self.url, VALID_URL_BASE)
            if not username:
                logger.warning(f"{AfreecaTV.__name__}: {self.url}: 直播间地址错误")
                return False

            channel_info = (await biliup.common.util.client.post(CHANNEL_API_URL, data={
                "bid": username,
                "bno": "",
                "type": "live",
                "pwd": "",
                "player_type": "html5",
                "stream_type": "common",
                "quality": QUALITIES[0],
                "mode": "landing",
                "from_api": 0,
            }, headers=self.fake_headers, timeout=5)).json()

            if channel_info["CHANNEL"]["RESULT"] == -6:
                logger.warning(f"{AfreecaTV.__name__}: {self.url}: 检测失败,请检查账号密码设置")
                return False

            if channel_info["CHANNEL"]["RESULT"] != 1:
                return False

            self.room_title = channel_info["CHANNEL"]["TITLE"]

            if is_check:
                return True

            aid_info = (await biliup.common.util.client.post(CHANNEL_API_URL, data={
                "bid": username,
                "bno": channel_info["CHANNEL"]["BNO"],
                "type": "aid",
                "pwd": "",
                "player_type": "html5",
                "stream_type": "common",
                "quality": QUALITIES[0],
                "mode": "landing",
                "from_api": 0,
            }, headers=self.fake_headers, timeout=5)).json()

            view_info = (await biliup.common.util.client.get(f'{channel_info["CHANNEL"]["RMD"]}/broad_stream_assign.html', params={
                "return_type": channel_info["CHANNEL"]["CDN"],
                "broad_key": f'{channel_info["CHANNEL"]["BNO"]}-common-{QUALITIES[0]}-hls'
            }, headers=self.fake_headers, timeout=5)).json()

            self.raw_stream_url = view_info["view_url"] + "?aid=" + aid_info["CHANNEL"]["AID"]
        except:
            logger.warning(f"{AfreecaTV.__name__}: {self.url}: 获取错误，本次跳过")
            return False

        return True


class AfreecaTVUtils:
    _cookie: Optional[Dict[str, str]] = None
    _cookie_expires = None

    @staticmethod
    def get_cookie() -> Optional[Dict[str, str]]:
        with NamedLock("AfreecaTV_cookie_get"):
            if not AfreecaTVUtils._cookie or AfreecaTVUtils._cookie_expires <= time.time():
                username = config.get('user', {}).get('afreecatv_username', '')
                password = config.get('user', {}).get('afreecatv_password', '')
                if not username or not password:
                    return None
                response = requests.post("https://login.afreecatv.com/app/LoginAction.php", data={
                    "szUid": username,
                    "szPassword": password,
                    "szWork": "login",
                    "szType": "json",
                    "isSaveId": "true",
                    "isSavePw": "true",
                    "isSaveJoin": "true",
                    "isLoginRetain": "Y",
                })
                if response.json()["RESULT"] != 1:
                    return None

                cookie_dict = response.cookies.get_dict()
                AfreecaTVUtils._cookie = {
                    "RDB": cookie_dict["RDB"],
                    "PdboxBbs": cookie_dict["PdboxBbs"],
                    "PdboxTicket": cookie_dict["PdboxTicket"],
                    "PdboxSaveTicket": cookie_dict["PdboxSaveTicket"],
                }
                AfreecaTVUtils._cookie_expires = time.time() + (7 * 24 * 60 * 60)

            return AfreecaTVUtils._cookie



# 抖音的弹幕录制参考了 https://github.com/LyzenX/DouyinLiveRecorder 和 https://github.com/YunzhiYike/live-tool
# 2023.07.14：KNaiFen：这部分代码参考了https://github.com/SmallPeaches/DanmakuRender
# 2024.06.22: 添加来自 https://github.com/hua0512/stream-rec 修改后的 webmssdk.js，以计算 signature

import gzip

import aiohttp
import json
from urllib.parse import unquote
from biliup.config import config
from .douyin_util.dy_pb2 import ChatMessage, PushFrame, Response
from biliup.plugins import match1
from google.protobuf import json_format
import logging

logger = logging.getLogger('biliup')

class Douyin:
    headers = {
        # 'user-agent': random_user_agent(),
        'Referer': 'https://live.douyin.com/',
        'Cookie': config.get('user', {}).get('douyin_cookie', '')
    }
    heartbeat = b':\x02hb'
    heartbeatInterval = 10

    @staticmethod
    async def get_ws_info(url, context):
        async with aiohttp.ClientSession() as session:
            from biliup.plugins.douyin import DouyinUtils
            from .douyin_util import DouyinDanmakuUtils
            Douyin.headers['user-agent'] = DouyinUtils.DOUYIN_USER_AGENT
            if "/user/" in url:
                async with session.get(url, headers=Douyin.headers, timeout=5) as resp:
                    user_page = await resp.text()
                    user_page_data = unquote(
                        user_page.split('<script id="RENDER_DATA" type="application/json">')[1].split('</script>')[0])
                    room_id = match1(user_page_data, r'"web_rid":"([^"]+)"')
            else:
                room_id = url.split('douyin.com/')[1].split('/')[0].split('?')[0]

            if room_id[0] == "+":
                room_id = room_id[1:]

            if "ttwid" not in Douyin.headers['Cookie']:
                Douyin.headers['Cookie'] = f'ttwid={DouyinUtils.get_ttwid()};{Douyin.headers["Cookie"]}'

            async with session.get(
                    DouyinUtils.build_request_url(f"https://live.douyin.com/webcast/room/web/enter/?web_rid={room_id}"),
                    headers=Douyin.headers, timeout=5) as resp:
                room_info = json.loads(await resp.text())['data']['data'][0]

            USER_UNIQUE_ID = DouyinDanmakuUtils.get_user_unique_id()
            VERSION_CODE = 180800 # https://lf-cdn-tos.bytescm.com/obj/static/webcast/douyin_live/7697.782665f8.js -> a.ry
            WEBCAST_SDK_VERSION = "1.0.14-beta.0" # https://lf-cdn-tos.bytescm.com/obj/static/webcast/douyin_live/7697.782665f8.js -> ee.VERSION
            # logger.info(f"user_unique_id: {USER_UNIQUE_ID}")
            sig_params = {
                "live_id": "1",
                "aid": "6383",
                "version_code": VERSION_CODE,
                "webcast_sdk_version": WEBCAST_SDK_VERSION,
                "room_id": room_info['id_str'],
                "sub_room_id": "",
                "sub_channel_id": "",
                "did_rule": "3",
                "user_unique_id": USER_UNIQUE_ID,
                "device_platform": "web",
                "device_type": "",
                "ac": "",
                "identity": "audience"
            }
            signature = DouyinDanmakuUtils.get_signature(DouyinDanmakuUtils.get_x_ms_stub(sig_params))
            # logger.info(f"signature: {signature}")
            webcast5_params = {
                "room_id": room_info['id_str'],
                "compress": 'gzip',
                # "app_name": "douyin_web",
                "version_code": VERSION_CODE,
                "webcast_sdk_version": WEBCAST_SDK_VERSION,
                # "update_version_code": "1.0.14-beta.0",
                # "cookie_enabled": "true",
                # "screen_width": "1920",
                # "screen_height": "1080",
                # "browser_online": "true",
                # "tz_name": "Asia/Shanghai",
                # "cursor": "t-1718899404570_r-1_d-1_u-1_h-7382616636258522175",
                # "internal_ext": "internal_src:dim|wss_push_room_id:7382580251462732598|wss_push_did:7344670681018189347|first_req_ms:1718899404493|fetch_time:1718899404570|seq:1|wss_info:0-1718899404570-0-0|wrds_v:7382616716703957597",
                # "host": "https://live.douyin.com",
                "live_id": "1",
                "did_rule": "3",
                # "endpoint": "live_pc",
                # "support_wrds": "1",
                "user_unique_id": USER_UNIQUE_ID,
                # "im_path": "/webcast/im/fetch/",
                "identity": "audience",
                # "need_persist_msg_count": "15",
                # "insert_task_id": "",
                # "live_reason": "",
                # "heartbeatDuration": "0",
                "signature": signature,
            }
            wss_url = f"wss://webcast5-ws-web-lf.douyin.com/webcast/im/push/v2/?{'&'.join([f'{k}={v}' for k, v in webcast5_params.items()])}"
            url = DouyinUtils.build_request_url(wss_url)
            return url, []

    @staticmethod
    def decode_msg(data):
        wss_package = PushFrame()
        wss_package.ParseFromString(data)
        log_id = wss_package.logId
        decompressed = gzip.decompress(wss_package.payload)
        payload_package = Response()
        payload_package.ParseFromString(decompressed)

        ack = None
        if payload_package.needAck:
            obj = PushFrame()
            obj.payloadType = 'ack'
            obj.logId = log_id
            obj.payloadType = payload_package.internalExt
            ack = obj.SerializeToString()

        msgs = []
        for msg in payload_package.messagesList:
            if msg.method == 'WebcastChatMessage':
                chat_message = ChatMessage()
                chat_message.ParseFromString(msg.payload)
                data = json_format.MessageToDict(chat_message, preserving_proto_field_name=True)
                # name = data['user']['nickName']
                content = data['content']
                # print(content)
                # msg_dict = {"time": now, "name": name, "content": content, "msg_type": "danmaku", "color": "ffffff"}
                msg_dict = {"content": content, "msg_type": "danmaku"}
                msgs.append(msg_dict)

        return msgs, ack



import os
from logging.handlers import TimedRotatingFileHandler
import time
import logging

class SafeRotatingFileHandler(TimedRotatingFileHandler):
    def __init__(self, filename, when='h', interval=1, backupCount=0, encoding=None, delay=False, utc=False):
        TimedRotatingFileHandler.__init__(self, filename, when, interval, backupCount, encoding, delay, utc)

    """
    Override doRollover
    lines commanded by "##" is changed by cc
    """

    def doRollover(self):
        """
        do a rollover; in this case, a date/time stamp is appended to the filename
        when the rollover happens.  However, you want the file to be named for the
        start of the interval, not the current time.  If there is a backup count,
        then we have to get a list of matching filenames, sort them and remove
        the one with the oldest suffix.

        Override,   1. if dfn not exist then do rename
                    2. _open with "a" model
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        # get the time that this sequence started at and make it a TimeTuple
        currentTime = int(time.time())
        dstNow = time.localtime(currentTime)[-1]
        t = self.rolloverAt - self.interval
        if self.utc:
            timeTuple = time.gmtime(t)
        else:
            timeTuple = time.localtime(t)
            dstThen = timeTuple[-1]
            if dstNow != dstThen:
                if dstNow:
                    addend = 3600
                else:
                    addend = -3600
                timeTuple = time.localtime(t + addend)
        dfn = self.baseFilename + "." + time.strftime(self.suffix, timeTuple)
        ##        if os.path.exists(dfn):
        ##            os.remove(dfn)

        # Issue 18940: A file may not have been created if delay is True.
        ##        if os.path.exists(self.baseFilename):
        if not os.path.exists(dfn) and os.path.exists(self.baseFilename):
            os.rename(self.baseFilename, dfn)
        if self.backupCount > 0:
            for s in self.getFilesToDelete():
                os.remove(s)
        if not self.delay:
            self.mode = "a"
            self.stream = self._open()
        newRolloverAt = self.computeRollover(currentTime)
        while newRolloverAt <= currentTime:
            newRolloverAt = newRolloverAt + self.interval
        # If DST changes and midnight or weekly rollover, adjust for this.
        if (self.when == 'MIDNIGHT' or self.when.startswith('W')) and not self.utc:
            dstAtRollover = time.localtime(newRolloverAt)[-1]
            if dstNow != dstAtRollover:
                if not dstNow:  # DST kicks in before next rollover, so we need to deduct an hour
                    addend = -3600
                else:  # DST bows out before next rollover, so we need to add an hour
                    addend = 3600
                newRolloverAt += addend
        self.rolloverAt = newRolloverAt

class DebugLevelFilter(logging.Filter):
    """
    A logging filter that blocks any log messages unless the console level is set to DEBUG
    """
    def filter(self, record):
        return logging.getLogger().isEnabledFor(logging.DEBUG)


import json
import logging
import os
import shutil
import subprocess
import time
from functools import reduce
from pathlib import Path
from typing import List

from biliup.config import config
from .app import event_manager, context
from .common.tools import NamedLock, processor
from .database.db import get_stream_info_by_filename, SessionLocal
from .downloader import biliup_download
from .engine.event import Event
from .engine.upload import UploadBase
from .uploader import upload, fmt_title_and_desc

PRE_DOWNLOAD = 'pre_download'
DOWNLOAD = 'download'
DOWNLOADED = 'downloaded'
UPLOAD = 'upload'
UPLOADED = 'uploaded'
logger = logging.getLogger('biliup')


# @event_manager.register(CHECK, block='Asynchronous3')


@event_manager.register(PRE_DOWNLOAD, block='Asynchronous1')
def pre_processor(name, url):
    if context['PluginInfo'].url_status[url] == 1:
        logger.debug(f'{name} 正在下载中，跳过下载')
        return
    logger.info(f'{name} - {url} 开播了准备下载')
    preprocessor = config['streamers'].get(name, {}).get('preprocessor')
    if preprocessor:
        processor(preprocessor, json.dumps({
            "name": name,
            "url": url,
            "start_time": int(time.time())
        }, ensure_ascii=False))
    yield Event(DOWNLOAD, (name, url))


@event_manager.register(DOWNLOAD, block='Asynchronous1')
def process(name, url):
    url_status = context['PluginInfo'].url_status
    # 下载开始
    try:
        url_status[url] = 1
        stream_info = biliup_download(name, url, config['streamers'][name].copy())
        yield Event(DOWNLOADED, (stream_info,))
    except Exception as e:
        logger.exception(f"下载错误: {name} - {e}")
    finally:
        # 下载结束
        url_status[url] = 0


@event_manager.register(DOWNLOADED, block='Asynchronous1')
def processed(stream_info):
    name = stream_info['name']
    url = stream_info['url']
    # 下载后处理 上传前处理
    downloaded_processor = config['streamers'].get(name, {}).get('downloaded_processor')
    if downloaded_processor:
        default_date = time.localtime()
        file_list = UploadBase.file_list(name)
        processor(downloaded_processor, json.dumps({
            "name": name,
            "url": url,
            "room_title": stream_info.get('title', name),
            "start_time": int(time.mktime(stream_info.get('date', default_date))),
            "end_time": int(time.mktime(stream_info.get('end_time', default_date))),
            "file_list": [file.video for file in file_list]
        }, ensure_ascii=False))
        # 后处理完成后重新扫描文件列表
    yield Event(UPLOAD, (stream_info,))


@event_manager.register(UPLOAD, block='Asynchronous2')
def process_upload(stream_info):
    url = stream_info['url']
    name = stream_info['name']
    url_upload_count = context['url_upload_count']
    # 永远不可能有两个同url的下载线程
    # 可能对同一个url同时发送两次上传事件
    with NamedLock(f"upload_count_{url}"):
        if context['url_upload_count'][url] > 0:
            return logger.debug(f'{url} 正在上传中，跳过')
        # from .handler import event_manager, UPLOAD
        # += 不是原子操作
        context['url_upload_count'][url] += 1
    # 上传开始
    try:
        file_list = UploadBase.file_list(name)
        if len(file_list) <= 0:
            logger.debug("无需上传")
            return

        # 上传延迟检测
        url_status = context['PluginInfo'].url_status
        delay = int(config.get('delay', 0))
        if delay:
            logger.info(f'{name} -> {url} {delay}s 后检测是否上传')
            time.sleep(delay)
            if url_status[url] == 1:
                # 上传延迟检测，启用的话会在一段时间后检测是否存在下载任务，若存在则跳过本次上传
                return logger.info(f'{name} -> {url} 存在下载任务, 跳过本次上传')

        if ("title" not in stream_info) or (not stream_info["title"]):  # 如果 data 中不存在标题, 说明下载信息已丢失, 则尝试从数据库获取
            with SessionLocal() as db:
                data, _ = fmt_title_and_desc({
                    **get_stream_info_by_filename(db, os.path.splitext(file_list[0].video)[0]),
                    "name": name})  # 如果 restart, data 中会缺失 name 项
            stream_info.update(data)
        filelist = upload(stream_info)
        if filelist:
            uploaded(name, stream_info.get('live_cover_path'), filelist)
    except Exception:
        logger.exception(f"上传错误: {name}")
    finally:
        # 上传结束
        # 有可能有两个同url的上传线程 保证计数正确
        with NamedLock(f'upload_count_{url}'):
            url_upload_count[url] -= 1


def uploaded(name, live_cover_path, data: List):
    # data = file_list
    post_processor = config['streamers'].get(name, {}).get("postprocessor", None)
    if post_processor is None:
        # 删除封面
        if live_cover_path is not None:
            UploadBase.remove_file(live_cover_path)
        return UploadBase.remove_filelist(data)

    file_list = []
    for i in data:
        file_list.append(i.video)
        if i.danmaku is not None:
            file_list.append(i.danmaku)

    for post_processor in post_processor:
        if post_processor == 'rm':
            # 删除封面
            if live_cover_path is not None:
                UploadBase.remove_file(live_cover_path)
            UploadBase.remove_filelist(data)
            continue
        if post_processor.get('mv'):
            for file in file_list:
                path = Path(file)
                dest = Path(post_processor['mv'])
                if not dest.is_dir():
                    dest.mkdir(parents=True, exist_ok=True)
                try:
                    shutil.move(path, dest / path.name)
                except Exception as e:
                    logger.exception(e)
                    continue
                logger.info(f"move to {(dest / path.name).absolute()}")
        if post_processor.get('run'):
            try:
                process_output = subprocess.check_output(
                    post_processor['run'], shell=True,
                    input=reduce(lambda x, y: x + str(Path(y).absolute()) + '\n', file_list, ''),
                    stderr=subprocess.STDOUT, text=True)
                logger.info(process_output.rstrip())
            except subprocess.CalledProcessError as e:
                logger.exception(e.output)
                continue



import json
from typing import Optional
from urllib.parse import unquote, urlparse, parse_qs, urlencode, urlunparse

import requests

from biliup.common.util import client
from . import logger, match1, random_user_agent
from biliup.config import config
from biliup.Danmaku import DanmakuClient
from ..common.tools import NamedLock
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www|m|live)\.)?douyin\.com')
class Douyin(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.douyin_danmaku = config.get('douyin_danmaku', False)
        self.fake_headers['user-agent'] = DouyinUtils.DOUYIN_USER_AGENT
        self.fake_headers['referer'] = "https://live.douyin.com/"
        self.fake_headers['cookie'] = config.get('user', {}).get('douyin_cookie', '')

    async def acheck_stream(self, is_check=False):
        if "/user/" in self.url:
            try:
                user_page = (await client.get(self.url, headers=self.fake_headers)).text
                user_page_data = unquote(
                    user_page.split('<script id="RENDER_DATA" type="application/json">')[1].split('</script>')[0])
                room_id = match1(user_page_data, r'"web_rid":"([^"]+)"')
                if room_id is None or not room_id:
                    logger.debug(f"{Douyin.__name__}: {self.url}: 未开播")
                    return False
            except (KeyError, IndexError):
                logger.warning(f"{Douyin.__name__}: {self.url}: 获取房间ID失败,请检查Cookie设置")
                return False
            except:
                logger.exception(f"{Douyin.__name__}: {self.url}: 获取房间ID失败")
                return False
        else:
            try:
                room_id = self.url.split('douyin.com/')[1].split('/')[0].split('?')[0]
                if not room_id:
                    raise
            except:
                logger.warning(f"{Douyin.__name__}: {self.url}: 直播间地址错误")
                return False

        if room_id[0] == "+":
            room_id = room_id[1:]
        try:
            if "ttwid" not in self.fake_headers['cookie']:
                self.fake_headers['Cookie'] = f'ttwid={DouyinUtils.get_ttwid()};{self.fake_headers["cookie"]}'
            page = (await client.get(
                DouyinUtils.build_request_url(f"https://live.douyin.com/webcast/room/web/enter/?web_rid={room_id}"),
                headers=self.fake_headers)).json()
            room_info = page.get('data').get('data')
            if room_info is None:
                logger.warning(f"{Douyin.__name__}: {self.url}: {page}")
                return False
            if len(room_info) > 0:
                room_info = room_info[0]
            else:
                room_info = {}
        except:
            logger.exception(f"{Douyin.__name__} - {self.url}: room_info 获取失败")
            return False

        try:
            if room_info.get('status') != 2:
                logger.debug(f"{Douyin.__name__}: {self.url}: 未开播")
                return False
        except:
            logger.exception(f"{Douyin.__name__} - {self.url}: 获取开播状态失败")
            return False

        try:
            stream_data = json.loads(room_info['stream_url']['live_core_sdk_data']['pull_data']['stream_data'])['data']
        except:
            logger.exception(f"{Douyin.__name__} - {self.url}: 加载清晰度失败")
            return False

        try:
            # 原画origin 蓝光uhd 超清hd 高清sd 标清ld 流畅md 仅音频ao
            quality_items = ['origin', 'uhd', 'hd', 'sd', 'ld', 'md']
            quality = config.get('douyin_quality', 'origin')
            if quality not in quality_items:
                quality = quality_items[0]

            # 如果没有这个画质则取相近的 优先低清晰度
            if quality not in stream_data:
                # 可选的清晰度 含自身
                optional_quality_items = [x for x in quality_items if x in stream_data.keys() or x == quality]
                # 自身在可选清晰度的位置
                optional_quality_index = optional_quality_items.index(quality)
                # 自身在所有清晰度的位置
                quality_index = quality_items.index(quality)
                # 高清晰度偏移
                quality_left_offset = None
                # 低清晰度偏移
                quality_right_offset = None

                if optional_quality_index + 1 < len(optional_quality_items):
                    quality_right_offset = quality_items.index(
                        optional_quality_items[optional_quality_index + 1]) - quality_index

                if optional_quality_index - 1 >= 0:
                    quality_left_offset = quality_index - quality_items.index(
                        optional_quality_items[optional_quality_index - 1])

                # 取相邻的清晰度
                if quality_right_offset <= quality_left_offset:
                    quality = optional_quality_items[optional_quality_index + 1]
                else:
                    quality = optional_quality_items[optional_quality_index - 1]

            protocol = 'hls' if config.get('douyin_protocol') == 'hls' else 'flv'
            # protocol = 'hls'
            self.raw_stream_url = stream_data[quality]['main'][protocol]
            self.room_title = room_info['title']
        except:
            logger.exception(f"{Douyin.__name__} - {self.url}: 寻找清晰度失败")
            return False
        return True

    def danmaku_init(self):
        if self.douyin_danmaku:
            try:
                import jsengine
                try:
                    jsengine.jsengine()
                    self.danmaku = DanmakuClient(self.url, self.gen_download_filename())
                except jsengine.exceptions.RuntimeError as e:
                    extra_msg = "如需录制抖音弹幕，"
                    logger.error(f"\n{e}\n{extra_msg}请至少安装一个 Javascript 解释器，如 pip install quickjs")
            except:
                pass


class DouyinUtils:
    # 抖音ttwid
    _douyin_ttwid: Optional[str] = None
    # DOUYIN_USER_AGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.159 Safari/537.36'
    DOUYIN_USER_AGENT = random_user_agent()
    DOUYIN_HTTP_HEADERS = {
        'User-Agent': DOUYIN_USER_AGENT
    }

    @staticmethod
    def get_ttwid() -> Optional[str]:
        with NamedLock("douyin_ttwid_get"):
            if not DouyinUtils._douyin_ttwid:
                page = requests.get("https://live.douyin.com/1-2-3-4-5-6-7-8-9-0", timeout=15)
                DouyinUtils._douyin_ttwid = page.cookies.get("ttwid")
            return DouyinUtils._douyin_ttwid

    @staticmethod
    def build_request_url(url: str) -> str:
        parsed_url = urlparse(url)
        existing_params = parse_qs(parsed_url.query)
        existing_params['aid'] = ['6383']
        existing_params['device_platform'] = ['web']
        existing_params['browser_language'] = ['zh-CN']
        existing_params['browser_platform'] = ['Win32']
        existing_params['browser_name'] = [DouyinUtils.DOUYIN_USER_AGENT.split('/')[0]]
        existing_params['browser_version'] = [DouyinUtils.DOUYIN_USER_AGENT.split(existing_params['browser_name'][0])[-1][1:]]
        new_query_string = urlencode(existing_params, doseq=True)
        new_url = urlunparse((
            parsed_url.scheme,
            parsed_url.netloc,
            parsed_url.path,
            parsed_url.params,
            new_query_string,
            parsed_url.fragment
        ))
        return new_url



import io
import random
import re
import socket
import subprocess
import time
from typing import AsyncGenerator, List
from urllib.parse import urlencode

import yt_dlp

from biliup.common.util import client
from biliup.config import config
from biliup.Danmaku import DanmakuClient
from . import logger
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase, BatchCheck

VALID_URL_BASE = r'(?:https?://)?(?:(?:www|go|m)\.)?twitch\.tv/(?P<id>[0-9_a-zA-Z]+)'
VALID_URL_VIDEOS = r'https?://(?:(?:www|go|m)\.)?twitch\.tv/(?P<id>[^/]+)/(?:videos|profile|clips)'
_CLIENT_ID = 'kimne78kx3ncx6brgo4mv6wki5h1ko'


@Plugin.download(regexp=VALID_URL_VIDEOS)
class TwitchVideos(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        DownloadBase.__init__(self, fname, url, suffix=suffix)
        self.is_download = True
        self.twitch_download_entry = None

    async def acheck_stream(self, is_check=False):
        while True:
            auth_token = TwitchUtils.get_auth_token()
            if auth_token:
                cookie = io.StringIO(f"""# Netscape HTTP Cookie File
.twitch.tv	TRUE	/	FALSE	0	auth-token	{auth_token}
""")
            else:
                cookie = None

            with yt_dlp.YoutubeDL({'download_archive': 'archive.txt', 'cookiefile': cookie}) as ydl:
                try:
                    info = ydl.extract_info(self.url, download=False, process=False)
                    for entry in info['entries']:
                        if ydl.in_download_archive(entry):
                            continue
                        if not is_check:
                            download_info = ydl.extract_info(entry['url'], download=False)
                            self.room_title = download_info['title']
                            self.raw_stream_url = download_info['url']
                            thumbnails = download_info.get('thumbnails')
                            if type(thumbnails) is list and len(thumbnails) > 0:
                                self.live_cover_url = thumbnails[len(thumbnails) - 1].get('url')
                            self.twitch_download_entry = entry
                        return True
                except Exception as e:
                    if 'Unauthorized' in str(e):
                        TwitchUtils.invalid_auth_token()
                        continue
                    else:
                        logger.warning(f"{self.url}：获取错误", exc_info=True)
                return False

    def download_success_callback(self):
        with yt_dlp.YoutubeDL({'download_archive': 'archive.txt'}) as ydl:
            ydl.record_download_archive(self.twitch_download_entry)


@Plugin.download(regexp=VALID_URL_BASE)
class Twitch(DownloadBase, BatchCheck):
    def __init__(self, fname, url, suffix='flv'):
        DownloadBase.__init__(self, fname, url, suffix=suffix)
        self.twitch_danmaku = config.get('twitch_danmaku', False)
        self.twitch_disable_ads = config.get('twitch_disable_ads', True)
        self.__proc = None

    async def acheck_stream(self, is_check=False):
        channel_name = re.match(VALID_URL_BASE, self.url).group('id').lower()
        user = (await TwitchUtils.post_gql({
            "query": '''
                query query($channel_name:String!) {
                    user(login: $channel_name){
                        stream {
                            id
                            title
                            type
                            previewImageURL(width: 0,height: 0)
                            playbackAccessToken(
                                params: {
                                    platform: "web",
                                    playerBackend: "mediaplayer",
                                    playerType: "site"
                                }
                            ) {
                                signature
                                value
                            }
                        }
                    }
                }
            ''',
            'variables': {'channel_name': channel_name}
        })).get('data', {}).get('user')
        if not user:
            logger.warning(f"{Twitch.__name__}: {self.url}: 获取错误", exc_info=True)
            return False
        elif not user['stream'] or user['stream']['type'] != 'live':
            return False

        self.room_title = user['stream']['title']
        self.live_cover_url = user['stream']['previewImageURL']
        if is_check:
            return True

        # https://github.com/biliup/biliup/issues/991
        if self.downloader == 'ffmpeg':
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(('localhost', 0))
                port = s.getsockname()[1]

            stream_shell = [
                "streamlink",
                "--player-external-http",  # 为外部程序提供流媒体数据
                "--player-external-http-port", str(port),  # 对外部输出流的端口
                "--player-external-http-interface", "localhost",
                # "--twitch-disable-ads",                     # 去广告，去掉、跳过嵌入的广告流
                # "--twitch-disable-hosting",               # 该参数从5.0起已被禁用
                "--twitch-disable-reruns",  # 如果该频道正在重放回放，不打开流
                self.url, "best"  # 流链接
            ]
            if self.twitch_disable_ads:  # 去广告，去掉、跳过嵌入的广告流
                stream_shell.insert(1, "--twitch-disable-ads")

            auth_token = TwitchUtils.get_auth_token()
            # 在设置且有效的情况下使用
            if auth_token:
                stream_shell.insert(1, f"--twitch-api-header=Authorization=OAuth {auth_token}")

            self.__proc = subprocess.Popen(stream_shell)
            self.raw_stream_url = f"http://localhost:{port}"
            i = 0
            while i < 5:
                if not (self.__proc.poll() is None):
                    return False
                time.sleep(1)
                i += 1
            return True
        else:
            query = {
                "player": "twitchweb",
                "p": random.randint(1000000, 10000000),
                "allow_source": "true",
                "allow_audio_only": "true",
                "allow_spectre": "false",
                'fast_bread': "true",
                'sig': user.get('stream').get('playbackAccessToken').get('signature'),
                'token': user.get('stream').get('playbackAccessToken').get('value'),
            }
            self.raw_stream_url = f'https://usher.ttvnw.net/api/channel/hls/{channel_name}.m3u8?{urlencode(query)}'
            return True

    @staticmethod
    async def abatch_check(check_urls: List[str]) -> AsyncGenerator[str, None]:
        ops = []
        for url in check_urls:
            channel_name = re.match(VALID_URL_BASE, url).group('id')
            op = {
                "query": '''
                    query query($login:String!) {
                        user(login: $login){
                            stream {
                              type
                            }
                        }
                    }
                ''',
                'variables': {'login': channel_name.lower()}
            }
            ops.append(op)
        gql = await TwitchUtils.post_gql(ops)
        for index, data in enumerate(gql):
            user = data.get('data', {}).get('user')
            if not user:
                logger.warning(f"{Twitch.__name__}: {check_urls[index]}: 获取错误")
                continue
            elif not user['stream'] or user['stream']['type'] != 'live':
                continue
            yield check_urls[index]

    def danmaku_init(self):
        if self.twitch_danmaku:
            self.danmaku = DanmakuClient(self.url, self.gen_download_filename())

    def close(self):
        try:
            if self.__proc is not None:
                self.__proc.terminate()
                self.__proc.wait(timeout=5)
        except subprocess.TimeoutExpired:
            self.__proc.kill()
        except:
            logger.exception(f'terminate {self.fname} failed')
        finally:
            self.__proc = None


class TwitchUtils:
    # Twitch已失效的auth_token
    _invalid_auth_token = None

    @staticmethod
    def get_auth_token():
        auth_token = config.get('user', {}).get('twitch_cookie')
        if TwitchUtils._invalid_auth_token == auth_token:
            return None
        return auth_token

    @staticmethod
    def invalid_auth_token():
        TwitchUtils._invalid_auth_token = config.get('user', {}).get('twitch_cookie')
        logger.warning("Twitch Cookie已失效请及时更换，后续操作将忽略Twitch Cookie")

    @staticmethod
    async def post_gql(ops):
        headers = {
            'Content-Type': 'text/plain;charset=UTF-8',
            'Client-ID': _CLIENT_ID,
        }
        auth_token = TwitchUtils.get_auth_token()
        if auth_token:
            headers['Authorization'] = f'OAuth {auth_token}'

        if isinstance(ops, list):
            limit = 30
            ops_list = [ops[i:i + limit] for i in range(0, len(ops), limit)]
            data = []
            for __ops in ops_list:
                __data = await TwitchUtils.__post_gql(headers, __ops)
                if __data: # 让检测不抛出异常
                    data.extend(__data)
            return data

        # 正常下载由上层方法处理
        return await TwitchUtils.__post_gql(headers, ops)

    @staticmethod
    async def __post_gql(headers, ops):
        try:
            _resp = await client.post(
                'https://gql.twitch.tv/gql',
                json=ops,
                headers=headers,
                timeout=15)
            _resp.raise_for_status()
            gql = _resp.json()
            if isinstance(gql, dict) and gql.get('error') == 'Unauthorized':
                TwitchUtils.invalid_auth_token()
                return await TwitchUtils.post_gql(ops)
            return gql
        except:
            logger.exception(f"Twitch - post_gql: {ops}")
        return {}


# Tencent is pleased to support the open source community by making Tars available.
#
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
#

import string
import struct

from .__packet import RequestPacket
from .__tars import TarsInputStream
from .__tars import TarsOutputStream
from .__util import util


class TarsUniPacket(object):
    def __init__(self):
        self.__mapa = util.mapclass(util.string, util.bytes)
        self.__mapv = util.mapclass(util.string, self.__mapa)
        self.__buffer = self.__mapv()
        self.__code = RequestPacket()

    # @property
    # def version(self):
    #     return self.__code.iVersion

    # @version.setter
    # def version(self, value):
    #     self.__code.iVersion = value

    @property
    def servant(self):
        return self.__code.sServantName

    @servant.setter
    def servant(self, value):
        self.__code.sServantName = value

    @property
    def func(self):
        return self.__code.sFuncName

    @func.setter
    def func(self, value):
        self.__code.sFuncName = value

    @property
    def requestid(self):
        return self.__code.iRequestId

    @requestid.setter
    def requestid(self, value):
        self.__code.iRequestId = value

    @property
    def result_code(self):
        if ("STATUS_RESULT_CODE" in self.__code.status) == False:
            return 0

        return string.atoi(self.__code.status["STATUS_RESULT_CODE"])

    @property
    def result_desc(self):
        if ("STATUS_RESULT_DESC" in self.__code.status) == False:
            return ''

        return self.__code.status["STATUS_RESULT_DESC"]

    def put(self, vtype, name, value):
        oos = TarsOutputStream()
        oos.write(vtype, 0, value)
        self.__buffer[name] = {vtype.__tars_class__: oos.getBuffer()}

    def get(self, vtype, name):
        if (name in self.__buffer) == False:
            raise Exception("UniAttribute not found key:%s,type:%s" %
                            (name, vtype.__tars_class__))

        t = self.__buffer[name]
        if (vtype.__tars_class__ in t) == False:
            raise Exception("UniAttribute not found type:" +
                            vtype.__tars_class__)

        o = TarsInputStream(t[vtype.__tars_class__])
        return o.read(vtype, 0, True)

    def encode(self):
        oos = TarsOutputStream()
        oos.write(self.__mapv, 0, self.__buffer)

        self.__code.iVersion = 2
        self.__code.sBuffer = oos.getBuffer()

        sos = TarsOutputStream()
        RequestPacket.writeTo(sos, self.__code)

        return struct.pack('!i', 4 + len(sos.getBuffer())) + sos.getBuffer()

    def decode(self, buf):
        ois = TarsInputStream(buf[4:])
        self.__code = RequestPacket.readFrom(ois)

        sis = TarsInputStream(self.__code.sBuffer)
        self.__buffer = sis.read(self.__mapv, 0, True)

    def clear(self):
        self.__code.__init__()

    def haskey(self, name):
        return name in self.__buffer



import asyncio

import httpx

# Set up for non-mainland China networks
HTTP_TIMEOUT = 15

client = httpx.AsyncClient(http2=True, follow_redirects=True, timeout=HTTP_TIMEOUT)
loop = asyncio.get_running_loop()



import hashlib
import time
from urllib.parse import parse_qs
from functools import lru_cache

from biliup.common.util import client
from biliup.config import config
from biliup.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from ..plugins import logger, match1, random_user_agent


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www|m)\.)?douyu\.com')
class Douyu(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self._room_id = match1(url, r'rid=(\d+)')
        self.douyu_danmaku = config.get('douyu_danmaku', False)

    async def acheck_stream(self, is_check=False):
        if len(self.url.split("douyu.com/")) < 2:
            logger.error(f"{self.plugin_msg}: 直播间地址错误")
            return False

        try:
            if not self._room_id:
                self._room_id = _get_real_rid(self.url)
        except:
            logger.exception(f"{self.plugin_msg}: 获取房间号错误")
            return False

        try:
            room_info = (
                await client.get(f"https://www.douyu.com/betard/{self._room_id}", headers=self.fake_headers)
            ).json()['room']
        except:
            logger.exception(f"{self.plugin_msg}: 获取直播间信息错误")
            return False

        if room_info['show_status'] != 1:
            logger.debug(f"{self.plugin_msg}: 未开播")
            return False
        if room_info['videoLoop'] != 0:
            logger.debug(f"{self.plugin_msg}: 正在放录播")
            return False
        if config.get('douyu_disable_interactive_game', False):
            gift_info = (
                await client.get(f"https://www.douyu.com/api/interactive/web/v2/list?rid={self._room_id}",
                                headers=self.fake_headers)
            ).json().get('data', {})
            if gift_info:
                logger.debug(f"{self.plugin_msg}: 正在运行互动游戏")
                return False
        self.room_title = room_info['room_name']

        if is_check:
            try:
                import jsengine
                try:
                    jsengine.jsengine()
                except jsengine.exceptions.RuntimeError as e:
                    extra_msg = "如需录制斗鱼直播，"
                    logger.error(f"\n{e}\n{extra_msg}请至少安装一个 Javascript 解释器，如 pip install quickjs")
                    return False
            except:
                logger.exception(f"{self.plugin_msg}: ")
                return False
            return True

        try:
            import jsengine
            ctx = jsengine.jsengine()
            js_enc = (
                await client.get(f'https://www.douyu.com/swf_api/homeH5Enc?rids={self._room_id}',
                                 headers=self.fake_headers)
            ).json()['data'][f'room{self._room_id}']
            js_enc = js_enc.replace('return eval', 'return [strc, vdwdae325w_64we];')

            sign_fun, sign_v = ctx.eval(f'{js_enc};ub98484234();')

            tt = str(int(time.time()))
            did = hashlib.md5(tt.encode('utf-8')).hexdigest()
            rb = hashlib.md5(f"{self._room_id}{did}{tt}{sign_v}".encode('utf-8')).hexdigest()
            sign_fun = sign_fun.rstrip(';').replace("CryptoJS.MD5(cb).toString()", f'"{rb}"')
            sign_fun += f'("{self._room_id}","{did}","{tt}");'

            params = parse_qs(ctx.eval(sign_fun))
        except:
            logger.exception(f"{self.plugin_msg}: 获取签名参数异常")
            return False

        params['cdn'] = config.get('douyucdn', 'tct-h5')
        params['cdn'] = config.get('douyu_cdn', params['cdn'])
        params['rate'] = config.get('douyu_rate', 0)

        try:
            live_data = await self.get_play_info(self._room_id, params)
            self.raw_stream_url = f"{live_data['rtmp_url']}/{live_data['rtmp_live']}"
        except:
            logger.exception(f"{self.plugin_msg}: ")
            return False

        return True

    def danmaku_init(self):
        if self.douyu_danmaku:
            self.danmaku = DanmakuClient(self.url, self.gen_download_filename())

    async def get_play_info(self, room_id, params):
        live_data = await client.post(
            f'https://www.douyu.com/lapi/live/getH5Play/{room_id}', headers=self.fake_headers, params=params)
        if not live_data.is_success:
            raise RuntimeError(live_data.text)
        live_data = live_data.json().get('data')
        if isinstance(live_data, dict):
            if not live_data['rtmp_cdn'].endswith('h5') or 'akm' in live_data['rtmp_cdn']:
                params['cdn'] = live_data['cdnsWithName'][-1]['cdn']
                return await self.get_play_info(room_id, params)
            return live_data
        raise RuntimeError(live_data)

@lru_cache(maxsize=None)
def _get_real_rid(url):
    import requests
    headers = {
        "user-agent": random_user_agent('mobile'),
    }
    rid = url.split('douyu.com/')[1].split('/')[0].split('?')[0] or match1(url, r'douyu.com/(\d+)')
    resp = requests.get(f"https://m.douyu.com/{rid}", headers=headers)
    real_rid = match1(resp.text, r'roomInfo":{"rid":(\d+)')
    return real_rid


import asyncio
import logging
import os
import re
import subprocess
import threading
import time
import shutil
from abc import ABC, abstractmethod
from typing import AsyncGenerator, List, Callable, Optional
from urllib.parse import urlparse

import requests
from requests.utils import DEFAULT_ACCEPT_ENCODING
from httpx import HTTPStatusError

from biliup.common.util import client, loop
from biliup.database.db import add_stream_info, SessionLocal, update_cover_path, update_room_title, update_file_list
from biliup.plugins import random_user_agent
import stream_gears
from PIL import Image

from biliup.config import config
from biliup.Danmaku import IDanmakuClient

logger = logging.getLogger('biliup')


class DownloadBase(ABC):
    def __init__(self, fname, url, suffix=None, opt_args=None):
        self.room_title = None
        if opt_args is None:
            opt_args = []
        self.fname = fname
        self.url = url
        # 录制后保存文件格式而非源流格式 对应原配置文件format 仅ffmpeg及streamlink生效
        if not suffix:
            logger.error(f'检测到suffix不存在，请补充后缀')
        else:
            self.suffix = suffix.lower()
        self.live_cover_path = None
        self.database_row_id = 0
        self.downloader = config.get('downloader', 'stream-gears')
        # ffmpeg.exe -i  http://vfile1.grtn.cn/2018/1542/0254/3368/154202543368.ssm/154202543368.m3u8
        # -c copy -bsf:a aac_adtstoasc -movflags +faststart output.mp4
        self.raw_stream_url = None

        # 主播单独传参会覆盖全局设置。例如新增了一个全局的filename_prefix参数，在下面添加self.filename_prefix = config.get('filename_prefix'),
        # 即可通过self.filename_prefix在下载或者上传时候传递主播单独的设置参数用于调用（如果该主播有设置单独参数，将会优先使用单独参数；如无，则会优先你用全局参数。）
        self.filename_prefix = config.get('filename_prefix')
        self.use_live_cover = config.get('use_live_cover', False)
        self.opt_args = opt_args
        self.live_cover_url = None
        self.fake_headers = {
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'accept-encoding': DEFAULT_ACCEPT_ENCODING,
            'accept-language': 'zh-CN,zh;q=0.8,en-US;q=0.5,en;q=0.3',
            'user-agent': random_user_agent(),
        }
        self.segment_time = config.get('segment_time', '01:00:00')
        self.file_size = config.get('file_size')

        # 是否是下载模式 跳过下播检测
        self.is_download = False

        # 分段后处理
        self.segment_processor = config.get('segment_processor')
        self.segment_processor_thread = []
        # 分段后处理并行
        self.segment_processor_parallel = config.get('segment_processor_parallel', False)

        # 弹幕客户端
        self.danmaku: Optional[IDanmakuClient] = None

        self.plugin_msg = f"{self.__class__.__name__} - {url}"

    @abstractmethod
    async def acheck_stream(self, is_check=False):
        # is_check 是否是检测模式 检测模式可以忽略只有下载时需要的耗时操作
        raise NotImplementedError()

    def download(self):
        logger.debug(f"{self.plugin_msg}: Start downloading {self.raw_stream_url}")
        if self.is_download:
            if not shutil.which("ffmpeg"):
                logger.error("未安装 FFMpeg 或不存在于 PATH 内")
                logger.debug("Current user's PATH is:" + os.getenv("PATH"))
                return False
            else:
                return self.ffmpeg_segment_download()

        parsed_url_path = urlparse(self.raw_stream_url).path
        if self.downloader == 'streamlink' or self.downloader == 'ffmpeg':
            if shutil.which("ffmpeg"):
                # streamlink无法处理flv,所以回退到ffmpeg
                if self.downloader == 'streamlink' and '.flv' not in parsed_url_path:
                    return self.ffmpeg_download(use_streamlink=True)
                else:
                    return self.ffmpeg_download()
            else:
                logger.error("未安装 FFMpeg 或不存在于 PATH 内，本次下载使用 stream-gears")
                logger.debug("Current user's PATH is:" + os.getenv("PATH"))

        if '.flv' in parsed_url_path:
            # 假定flv流
            self.suffix = 'flv'
        else:
            # 其他流stream_gears会按hls保存为ts
            self.suffix = 'ts'
        stream_gears_download(self.raw_stream_url, self.fake_headers, self.gen_download_filename(),
                              self.segment_time,
                              self.file_size,
                              lambda file_name: self.__download_segment_callback(file_name))
        return True

    def ffmpeg_segment_download(self):
        # TODO 无日志
        # , '-report'
        # ffmpeg 输入参数
        input_args = [
            '-loglevel', 'quiet', '-y'
        ]
        # ffmpeg 输出参数
        output_args = [
            '-bsf:a', 'aac_adtstoasc'
        ]
        input_args += ['-headers', ''.join('%s: %s\r\n' % x for x in self.fake_headers.items()),
                       '-rw_timeout', '20000000']
        if '.m3u8' in urlparse(self.raw_stream_url).path:
            input_args += ['-max_reload', '1000']

        input_args += ['-i', self.raw_stream_url]

        output_args += ['-f', 'segment']
        # output_args += ['-segment_format', self.suffix]
        output_args += ['-segment_list', 'pipe:1']
        output_args += ['-segment_list_type', 'flat']
        output_args += ['-reset_timestamps', '1']
        # output_args += ['-strftime', '1']
        if self.segment_time:
            output_args += ['-segment_time', self.segment_time]
        else:
            # 避免适配两套
            output_args += ['-segment_time', '9999:00:00']

        output_args += ['-c', 'copy']
        output_args += self.opt_args
        file_name = self.gen_download_filename(is_fmt=True)
        args = ['ffmpeg', *input_args, *output_args, f'{file_name}_%d.{self.suffix}']
        with subprocess.Popen(args, stdin=subprocess.DEVNULL, stdout=subprocess.PIPE,
                              stderr=subprocess.DEVNULL) as proc:
            for line in iter(proc.stdout.readline, b''):  # b'\n'-separated lines
                try:
                    ffmpeg_file_name = line.rstrip().decode(errors='ignore')
                    time.sleep(1)
                    # 文件重命名
                    self.download_file_rename(ffmpeg_file_name, f'{file_name}.{self.suffix}')
                    self.__download_segment_callback(f'{file_name}.{self.suffix}')
                    file_name = self.gen_download_filename(is_fmt=True)
                except:
                    logger.error(f'分段事件失败：{self.__class__.__name__} - {self.fname}', exc_info=True)

        return proc.returncode == 0

    def ffmpeg_download(self, use_streamlink=False):
        # streamlink进程
        streamlink_proc = None
        # updatedFileList = False
        try:
            # 文件名不含后戳
            fmt_file_name = self.gen_download_filename(is_fmt=True)
            # ffmpeg 输入参数
            input_args = []
            # ffmpeg 输出参数
            output_args = [
                '-bsf:a', 'aac_adtstoasc',
            ]
            if use_streamlink:
                streamlink_cmd = [
                    'streamlink',
                    '--stream-segment-threads', '3',
                    '--hls-playlist-reload-attempts', '1',
                    '--http-header',
                    ';'.join([f'{key}={value}' for key, value in self.fake_headers.items()]),
                    self.raw_stream_url,
                    'best',
                    '-O'
                ]
                streamlink_proc = subprocess.Popen(streamlink_cmd, stdout=subprocess.PIPE)
                input_uri = 'pipe:0'
            else:
                input_args += ['-headers', ''.join('%s: %s\r\n' % x for x in self.fake_headers.items()),
                               '-rw_timeout', '20000000']
                if '.m3u8' in urlparse(self.raw_stream_url).path:
                    input_args += ['-max_reload', '1000']
                input_uri = self.raw_stream_url

            input_args += ['-i', input_uri]

            if self.segment_time:
                output_args += ['-to', self.segment_time]
            if self.file_size:
                output_args += ['-fs', str(self.file_size)]

            output_args += self.opt_args

            if self.suffix.lower() == 'ts':
                output_args += ['-f', 'mpegts']
            elif self.suffix.lower() == 'mkv':
                output_args += ['-f', 'matroska']
            else:
                output_args += ['-f', self.suffix]
            args = ['ffmpeg', '-y', *input_args, *output_args, '-c', 'copy',
                    f'{fmt_file_name}.{self.suffix}.part']
            with subprocess.Popen(args, stdin=subprocess.DEVNULL if not streamlink_proc else streamlink_proc.stdout,
                                  stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as proc:
                # with SessionLocal() as db:
                #     update_file_list(db, self.database_row_id, fmt_file_name)
                #     updatedFileList = True
                for line in iter(proc.stdout.readline, b''):  # b'\n'-separated lines
                    decode_line = line.rstrip().decode(errors='ignore')
                    print(decode_line)
                    logger.debug(decode_line)

            if proc.returncode == 0:
                # 文件重命名
                self.download_file_rename(f'{fmt_file_name}.{self.suffix}.part', f'{fmt_file_name}.{self.suffix}')
                # 触发分段事件
                self.__download_segment_callback(f'{fmt_file_name}.{self.suffix}')
                return True
            else:
                return False
        # except:
        #     if updatedFileList:
        #         with SessionLocal() as db:
        #             delete_file_list(db, self.database_row_id, None)
        finally:
            try:
                if streamlink_proc:
                    streamlink_proc.terminate()
                    streamlink_proc.wait(timeout=5)
            except subprocess.TimeoutExpired:
                streamlink_proc.kill()
            except:
                logger.exception(f'terminate {self.fname} failed')

    def __download_segment_callback(self, file_name: str):
        """
        分段后触发返回含后戳的文件名
        """
        exclude_ext_file_name = os.path.splitext(file_name)[0]
        danmaku_file_name = os.path.splitext(file_name)[0] + '.xml'
        if self.danmaku:
            self.danmaku.save(danmaku_file_name)

        def x():
            # 将文件名和直播标题存储到数据库
            with SessionLocal() as db:
                update_file_list(db, self.database_row_id, file_name)
            if self.segment_processor:
                try:
                    if not self.segment_processor_parallel and prev_thread:
                        prev_thread.join()
                    from biliup.common.tools import processor
                    data = os.path.abspath(file_name)
                    if os.path.exists(danmaku_file_name):
                        data += f'\n{os.path.abspath(danmaku_file_name)}'
                    processor(self.segment_processor, data)
                except:
                    logger.warning(f'执行后处理失败：{self.__class__.__name__} - {self.fname}', exc_info=True)

        thread = threading.Thread(target=x, daemon=True, name=f"segment_processor_{exclude_ext_file_name}")
        prev_thread = self.segment_processor_thread[-1] if self.segment_processor_thread else None
        self.segment_processor_thread.append(thread)
        thread.start()

    def download_success_callback(self):
        pass

    def run(self):
        try:
            if not asyncio.run_coroutine_threadsafe(self.acheck_stream(), loop).result():
                return False
            with SessionLocal() as db:
                update_room_title(db, self.database_row_id, self.room_title)
            self.danmaku_init()
            if self.danmaku:
                self.danmaku.start()
            retval = self.download()
            return retval
        finally:
            if self.danmaku:
                self.danmaku.stop()
                self.danmaku = None

    def start(self):
        logger.info(f'开始下载: {self.__class__.__name__} - {self.fname}')
        # 开始时间
        start_time = time.localtime()
        # 结束时间
        end_time = None

        with SessionLocal() as db:
            self.database_row_id = add_stream_info(db, self.fname, self.url, start_time)  # 返回数据库中此行记录的 id
        ret = True
        while ret:
            # 下载结果
            try:
                ret = self.run()
            except Exception:
                logger.warning(f'下载失败: {self.__class__.__name__} - {self.fname}', exc_info=True)
            finally:
                self.close()

            # 下载模式跳过下播延迟检测
            if self.is_download:
                break

            # 最后一次下载完成时间
            end_time = time.localtime()

        self.download_cover(
            time.strftime(self.gen_download_filename().encode("unicode-escape").decode(), end_time if end_time else time.localtime()
                           ).encode().decode("unicode-escape"))
        # 更新数据库中封面存储路径
        with SessionLocal() as db:
            update_cover_path(db, self.database_row_id, self.live_cover_path)

        for thread in self.segment_processor_thread:
            if thread.is_alive():
                logger.info(f'等待分段后处理完成: {self.__class__.__name__} - {self.fname} - {thread.name}')
                thread.join()
        if (self.is_download and ret) or not self.is_download:
            self.download_success_callback()
        # self.segment_processor_thread
        logger.info(f'退出下载: {self.__class__.__name__} - {self.fname}')
        stream_info = {
            'name': self.fname,
            'url': self.url,
            'title': self.room_title,
            'date': start_time,
            'end_time': end_time if end_time else time.localtime(),
            'live_cover_path': self.live_cover_path,
            'is_download': self.is_download,
        }
        return stream_info

    def download_cover(self, fmtname):
        # 获取封面
        if self.use_live_cover and self.live_cover_url is not None:
            try:
                save_dir = f'cover/{self.__class__.__name__}/{self.fname}/'
                if not os.path.exists(save_dir):
                    os.makedirs(save_dir)

                url_path = urlparse(self.live_cover_url).path
                suffix = None
                if '.jpg' in url_path:
                    suffix = 'jpg'
                elif '.png' in url_path:
                    suffix = 'png'
                elif '.webp' in url_path:
                    suffix = 'webp'

                if suffix:
                    live_cover_path = f'{save_dir}{fmtname}.{suffix}'
                    if os.path.exists(live_cover_path):
                        self.live_cover_path = live_cover_path
                    else:
                        response = requests.get(self.live_cover_url, headers=self.fake_headers, timeout=30)
                        with open(live_cover_path, 'wb') as f:
                            f.write(response.content)

                    if suffix == 'webp':
                        with Image.open(live_cover_path) as img:
                            img = img.convert('RGB')
                            img.save(f'{save_dir}{fmtname}.jpg', format='JPEG')
                        os.remove(live_cover_path)
                        live_cover_path = f'{save_dir}{fmtname}.jpg'

                    self.live_cover_path = live_cover_path
                    logger.info(
                        f'封面下载成功：{self.__class__.__name__} - {self.fname}：{os.path.abspath(self.live_cover_path)}')
                else:
                    logger.warning(
                        f'封面下载失败：{self.__class__.__name__} - {self.fname}：封面格式不支持：{self.live_cover_url}')
            except:
                logger.exception(f'封面下载失败：{self.__class__.__name__} - {self.fname}')

    async def acheck_url_healthy(self, url):
        async def __client_get(url, stream: bool = False):
            client.headers.update(self.fake_headers)
            if stream:
                async with client.stream("GET", url, timeout=60, follow_redirects=False) as response:
                    pass
            else:
                response = await client.get(url)
            if response.status_code not in (301, 302):
                response.raise_for_status()
            return response

        try:
            if '.m3u8' in url:
                r = await __client_get(url)
                import m3u8
                m3u8_obj = m3u8.loads(r.text)
                if m3u8_obj.is_variant:
                    url = m3u8_obj.playlists[0].uri
                    logger.info(f'stream url: {url}')
                    r = await __client_get(url)
            else: # 处理 Flv
                r = await __client_get(url, stream=True)
                if r.headers.get('Location'):
                    url = r.headers['Location']
                    logger.info(f'stream url: {url}')
                    r = await __client_get(url, stream=True)
            if r.status_code == 200:
                return url
        except HTTPStatusError as e:
            logger.error(f'url {url}: status_code-{e.response.status_code}')
        except:
            logger.exception(f'url {url}: ')
        return None

    def gen_download_filename(self, is_fmt=False):
        if self.filename_prefix:  # 判断是否存在自定义录播命名设置
            filename = (self.filename_prefix.format(streamer=self.fname, title=self.room_title).encode(
                'unicode-escape').decode()).encode().decode("unicode-escape")
        else:
            filename = f'{self.fname}%Y-%m-%dT%H_%M_%S'
        filename = get_valid_filename(filename)
        if is_fmt:
            file_time = time.time()
            while True:
                fmt_file_name = time.strftime(filename.encode("unicode-escape").decode(),
                                              time.localtime(file_time)).encode().decode("unicode-escape")
                if os.path.exists(f"{fmt_file_name}.{self.suffix}"):
                    file_time += 1
                else:
                    return fmt_file_name
        else:
            return filename

    @staticmethod
    def download_file_rename(old_file_name, file_name):
        try:
            os.rename(old_file_name, file_name)
            logger.info(f'更名 {old_file_name} 为 {file_name}')
        except:
            logger.error(f'更名 {old_file_name} 为 {file_name} 失败', exc_info=True)

    def danmaku_init(self):
        pass

    def close(self):
        pass


def stream_gears_download(url, headers, file_name, segment_time=None, file_size=None,
                          file_name_callback: Callable[[str], None] = None):
    class Segment:
        pass

    segment = Segment()
    if segment_time:
        seg_time = segment_time.split(':')
        # print(int(seg_time[0]) * 60 * 60 + int(seg_time[1]) * 60 + int(seg_time[2]))
        segment.time = int(seg_time[0]) * 60 * 60 + int(seg_time[1]) * 60 + int(seg_time[2])
    if file_size:
        segment.size = file_size
    if file_size is None and segment_time is None:
        segment.size = 8 * 1024 * 1024 * 1024
    # FIXME: 下载时如出现403，这里不会回到上层方法获取新链接
    if file_name_callback:
        stream_gears.download_with_callback(
            url,
            headers,
            file_name,
            segment,
            file_name_callback
        )
    else:
        stream_gears.download(
            url,
            headers,
            file_name,
            segment,
        )


def get_valid_filename(name):
    """
    Return the given string converted to a string that can be used for a clean
    filename. Remove leading and trailing spaces; convert other spaces to
    underscores; and remove anything that is not an alphanumeric, dash,
    underscore, or dot.
    # >>> get_valid_filename("john's portrait in 2004.jpg")
    >>> get_valid_filename("{self.fname}%Y-%m-%dT%H_%M_%S")
    '{self.fname}%Y-%m-%dT%H_%M_%S'
    """
    # s = str(name).strip().replace(" ", "_") #因为有些人会在主播名中间加入空格，为了避免和录播完毕自动改名冲突，所以注释掉
    s = re.sub(r"(?u)[^-\w.%{}\[\]【】「」（）・°\s]", "", str(name))
    if s in {"", ".", ".."}:
        raise RuntimeError("Could not derive file name from '%s'" % name)
    return s


class BatchCheck(ABC):
    @staticmethod
    @abstractmethod
    async def abatch_check(check_urls: List[str]) -> AsyncGenerator[str, None]:
        """
        批量检测直播或下载状态
        返回的是url_list
        """



import json
import pathlib
import shutil
import os
from collections import UserDict
from sqlalchemy import select


try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib


class Config(UserDict):
    def load_cookies(self, file='cookies.json'):
        if not os.path.exists(file):
            raise FileNotFoundError(file)

        if self.data.get('user') is None:
            self.data['user'] = {'cookies': {}}
        if self.data["user"].get("cookies") is None:
            self.data['user']['cookies'] = {}
        with open(file, encoding='utf-8') as stream:
            s = json.load(stream)
            for i in s["cookie_info"]["cookies"]:
                name = i["name"]
                self.data["user"]["cookies"][name] = i["value"]
            self.data["user"]["access_token"] = s["token_info"]["access_token"]

    def load_from_db(self, db):
        from biliup.database.models import Configuration, LiveStreamers
        context = {
            'url_upload_count': self.data.get('url_upload_count', {}),
            'upload_filename': self.data.get('upload_filename', []),
            'PluginInfo': self.data.get('PluginInfo')
        }

        for con in db.execute(select(Configuration.value).where(Configuration.key == 'config')):
            self.data = json.loads(con.value)
        self.data.update(context)
        self['streamers'] = {}
        for ls in db.scalars(select(LiveStreamers)):
            self['streamers'][ls.remark] = {
                k: v for k, v in ls.as_dict().items() if v and (k not in ['upload_streamers_id', 'id', 'remark'])}
            # self['streamers'][ls.remark].pop('upload_streamers')
            if ls.upload_streamers_id:
                self['streamers'][ls.remark].update({
                    k: v for k, v in ls.uploadstreamers.as_dict().items() if v and k not in ['id', 'template_name']})
                if self['streamers'][ls.remark].get('uploader') is None:
                    self['streamers'][ls.remark]['uploader'] = 'biliup-rs'
            # if self['streamers'][ls.remark].get('tags'):
            #     self['streamers'][ls.remark]['tags'] = self['streamers'][ls.remark]['tags']
        # for us in UploadStreamers.select():
        #     config.data[con.key] = con.value

    def save_to_db(self, db):
        from biliup.database.models import Configuration, LiveStreamers, UploadStreamers
        for k, v in self['streamers'].items():
            us = UploadStreamers(**UploadStreamers.filter_parameters(
                {"template_name": k, "tags": v.pop('tags', [k]), ** v}))
            db.add(us)
            db.flush()
            url = v.pop('url')
            urls = url if isinstance(url, list) else [url]  # 兼容 url 输入字符串和列表
            for url in urls:
                ls = LiveStreamers(**LiveStreamers.filter_parameters(
                    {"upload_streamers_id": us.id, "remark": k, "url": url, ** v}))
                db.add(ls)
        del self['streamers']
        configuration = Configuration(key='config', value=json.dumps(self.data))
        db.add(configuration)
        db.commit()

    def load(self, file):
        import yaml
        if file is None:
            if pathlib.Path('config.yaml').exists():
                file = open('config.yaml', 'rb')
            elif pathlib.Path('config.toml').exists():
                self.data['toml'] = True
                file = open('config.toml', "rb")
            else:
                raise FileNotFoundError('未找到配置文件，请先创建配置文件')
        with file as stream:
            if file.name.endswith('.toml'):
                self.data = tomllib.load(stream)
            else:
                self.data = yaml.load(stream, Loader=yaml.FullLoader)

    def create_without_config_input(self, file):
        import yaml
        if file is None:
            if pathlib.Path('config.toml').exists():
                file = open('config.toml', 'rb')
            elif pathlib.Path('config.yaml').exists():
                file = open('config.yaml', encoding='utf-8')
            else:
                try:
                    from importlib.resources import files
                except ImportError:
                    from importlib_resources import files
                shutil.copy(files("biliup.web").joinpath('public/config.toml'), '.')
                file = open('config.toml', 'rb')

            # else:
            #     try:
            #         from importlib.resources import files
            #     except ImportError:
            #         # Try backported to PY<37 `importlib_resources`.
            #         from importlib_resources import files
            #     shutil.copy(files("biliup.web").joinpath('public/config.yaml'), 'common')
            #     file = open('config.yaml', encoding='utf-8')

        with file as stream:
            if file.name.endswith('.toml'):
                self.data = tomllib.load(stream)
                self.data['toml'] = True
            else:
                self.data = yaml.load(stream, Loader=yaml.FullLoader)

    def save(self):
        if self.data.get('toml'):
            import tomli_w
            with open('config.toml', 'rb') as stream:
                old_data = tomllib.load(stream)
                old_data["lines"] = self.data["lines"]
                old_data["threads"] = self.data["threads"]
                old_data["streamers"] = self.data["streamers"]
            with open('config.toml', 'wb') as stream:
                tomli_w.dump(old_data, stream)
        else:
            import yaml
            with open('config.yaml', 'w+', encoding='utf-8') as stream:
                old_data = yaml.load(stream, Loader=yaml.FullLoader)
                old_data["user"]["cookies"] = self.data["user"]["cookies"]
                old_data["user"]["access_token"] = self.data["user"]["access_token"]
                old_data["lines"] = self.data["lines"]
                old_data["threads"] = self.data["threads"]
                old_data["streamers"] = self.data["streamers"]
                yaml.dump(old_data, stream, default_flow_style=False, allow_unicode=True)

    def dump(self, file):
        """ 转储到旧版配置文件 """
        if not file:
            file = 'config.toml'
        if os.path.exists(file):
            from datetime import datetime
            import logging
            logger = logging.getLogger('biliup')
            new_name = f'{file}.backup.{datetime.now().strftime("%Y%m%d%H%M%S")}'
            logger.info(f"{file} 文件已存在，已将原文件重命名为 {new_name}")
            os.rename(file, new_name)
        exclude_keys = ['PluginInfo', 'upload_filename', 'url_upload_count']
        temp = {k: v for k, v in self.data.items() if k not in exclude_keys}
        if self.data.get('yaml') or file.endswith(".yaml"):
            import yaml
            with open(file, 'w+', encoding='utf-8') as stream:
                yaml.dump(temp, stream, default_flow_style=False, allow_unicode=True)
        else:
            import tomli_w
            # print(config)
            with open(file, 'wb') as stream:
                tomli_w.dump(temp, stream)
        return file


config = Config()



import random
import time
from . import enc
from base64 import urlsafe_b64encode as b64enc
from urllib.parse import quote


def _header(video_id, channel_id) -> str:
    S1_3 = enc.rs(1, video_id)
    S1_5 = enc.rs(1, channel_id) + enc.rs(2, video_id)
    S1 = enc.rs(3, S1_3) + enc.rs(5, S1_5)
    S3 = enc.rs(48687757, enc.rs(1, video_id))
    header_replay = enc.rs(1, S1) + enc.rs(3, S3) + enc.nm(4, 1)
    return b64enc(header_replay)


def _build(video_id, channel_id, ts1, ts2, ts3, ts4, ts5, topchat_only) -> str:
    chattype = 4 if topchat_only else 1

    b1 = enc.nm(1, 0)
    b2 = enc.nm(2, 0)
    b3 = enc.nm(3, 0)
    b4 = enc.nm(4, 0)
    b7 = enc.rs(7, "")
    b8 = enc.nm(8, 0)
    b9 = enc.rs(9, "")
    timestamp2 = enc.nm(10, ts2)
    b11 = enc.nm(11, 3)
    b15 = enc.nm(15, 0)

    header = enc.rs(3, _header(video_id, channel_id))
    timestamp1 = enc.nm(5, ts1)
    s6 = enc.nm(6, 0)
    s7 = enc.nm(7, 0)
    s8 = enc.nm(8, 1)
    body = enc.rs(9, b"".join((b1, b2, b3, b4, b7, b8, b9, timestamp2, b11, b15)))
    timestamp3 = enc.nm(10, ts3)
    timestamp4 = enc.nm(11, ts4)
    s13 = enc.nm(13, chattype)
    chattype = enc.rs(16, enc.nm(1, chattype))
    s17 = enc.nm(17, 0)
    str19 = enc.rs(19, enc.nm(1, 0))
    timestamp5 = enc.nm(20, ts5)
    entity = b"".join(
        (
            header,
            timestamp1,
            s6,
            s7,
            s8,
            body,
            timestamp3,
            timestamp4,
            s13,
            chattype,
            s17,
            str19,
            timestamp5,
        )
    )
    continuation = enc.rs(119693434, entity)
    return quote(b64enc(continuation).decode())


def _times(past_sec):
    n = int(time.time())
    _ts1 = n - random.uniform(0, 1 * 3)
    _ts2 = n - random.uniform(0.01, 0.99)
    _ts3 = n - past_sec + random.uniform(0, 1)
    _ts4 = n - random.uniform(10 * 60, 60 * 60)
    _ts5 = n - random.uniform(0.01, 0.99)
    return list(map(lambda x: int(x * 1000000), [_ts1, _ts2, _ts3, _ts4, _ts5]))


def getparam(video_id, channel_id, past_sec=0, topchat_only=False) -> str:
    """
    Parameter
    ---------
    past_sec : int
        seconds to load past chat data
    topchat_only : bool
        if True, fetch only 'top chat'
    """
    return _build(video_id, channel_id, *_times(past_sec), topchat_only)



import copy
import os
import shutil
from typing import Optional

import yt_dlp

from yt_dlp import DownloadError
from yt_dlp.utils import DateRange
from biliup.config import config
from ..engine.decorators import Plugin
from . import logger
from ..engine.download import DownloadBase

VALID_URL_BASE = r'https?://(?:(?:www|m)\.)?youtube\.com/(?P<id>.*?)\??(.*?)'


@Plugin.download(regexp=VALID_URL_BASE)
class Youtube(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.ytb_danmaku = config.get('ytb_danmaku', False)
        self.youtube_cookie = config.get('user', {}).get('youtube_cookie')
        self.youtube_prefer_vcodec = config.get('youtube_prefer_vcodec')
        self.youtube_prefer_acodec = config.get('youtube_prefer_acodec')
        self.youtube_max_resolution = config.get('youtube_max_resolution')
        self.youtube_max_videosize = config.get('youtube_max_videosize')
        self.youtube_before_date = config.get('youtube_before_date')
        self.youtube_after_date = config.get('youtube_after_date')
        self.youtube_enable_download_live = config.get('youtube_enable_download_live', True)
        self.youtube_enable_download_playback = config.get('youtube_enable_download_playback', True)
        # 需要下载的 url
        self.download_url = None

    async def acheck_stream(self, is_check=False):
        with yt_dlp.YoutubeDL({
            'download_archive': 'archive.txt',
            'cookiefile': self.youtube_cookie,
            'ignoreerrors': True,
            'extractor_retries': 0,
        }) as ydl:
            # 获取信息的时候不要过滤
            ydl_archive = copy.deepcopy(ydl.archive)
            ydl.archive = set()
            if self.download_url is not None:
                # 直播在重试的时候特别处理
                info = ydl.extract_info(self.download_url, download=False)
            else:
                info = ydl.extract_info(self.url, download=False, process=False)
            if type(info) is not dict:
                logger.warning(f"{Youtube.__name__}: {self.url}: 获取错误")
                return False

            cache = KVFileStore(f"./cache/youtube/{self.fname}.txt")

            def loop_entries(entrie):
                if type(entrie) is not dict:
                    return None
                elif entrie.get('_type') == 'playlist':
                    # 播放列表递归
                    for e in entrie.get('entries'):
                        le = loop_entries(e)
                        if type(le) is dict:
                            return le
                        elif le == "stop":
                            return None
                elif type(entrie) is dict:
                    # is_upcoming 等待开播 is_live 直播中 was_live结束直播(回放)
                    if entrie.get('live_status') == 'is_upcoming':
                        return None
                    elif entrie.get('live_status') == 'is_live':
                        # 未开启直播下载忽略
                        if not self.youtube_enable_download_live:
                            return None
                    elif entrie.get('live_status') == 'was_live':
                        # 未开启回放下载忽略
                        if not self.youtube_enable_download_playback:
                            return None

                    # 检测是否已下载
                    if ydl._make_archive_id(entrie) in ydl_archive:
                        # 如果已下载但是还在直播则不算下载
                        if entrie.get('live_status') != 'is_live':
                            return None

                    upload_date = cache.query(entrie.get('id'))
                    if upload_date is None:
                        if entrie.get('upload_date') is not None:
                            upload_date = entrie['upload_date']
                        else:
                            entrie = ydl.extract_info(entrie.get('url'), download=False, process=False)
                            if type(entrie) is dict and entrie.get('upload_date') is not None:
                                upload_date = entrie['upload_date']

                        # 时间是必然存在的如果不存在说明出了问题 暂时跳过
                        if upload_date is None:
                            return None
                        else:
                            cache.add(entrie.get('id'), upload_date)

                    if self.youtube_after_date is not None and upload_date < self.youtube_after_date:
                        return 'stop'

                    # 检测时间范围
                    if upload_date not in DateRange(self.youtube_after_date, self.youtube_before_date):
                        return None

                    return entrie
                return None

            download_entry: Optional[dict] = loop_entries(info)
            if type(download_entry) is dict:
                if download_entry.get('live_status') == 'is_live':
                    self.is_download = False
                else:
                    self.is_download = True
                if not is_check:
                    if download_entry.get('_type') == 'url':
                        download_entry = ydl.extract_info(download_entry.get('url'), download=False, process=False)
                    self.room_title = download_entry.get('title')
                    self.live_cover_url = download_entry.get('thumbnail')
                    self.download_url = download_entry.get('webpage_url')
                return True
            else:
                return False

    def download(self):
        filename = self.gen_download_filename(is_fmt=True)
        # ydl下载的文件在下载失败时不可控
        # 临时存储在其他地方
        download_dir = f'./cache/temp/youtube/{filename}'
        try:
            ydl_opts = {
                'outtmpl': f'{download_dir}/{filename}.%(ext)s',
                'cookiefile': self.youtube_cookie,
                'break_on_reject': True,
                'download_archive': 'archive.txt',
                'format': 'bestvideo',
                # 'proxy': proxyUrl,
            }

            if self.youtube_prefer_vcodec is not None:
                ydl_opts['format'] += f"[vcodec~='^({self.youtube_prefer_vcodec})']"
            if self.youtube_max_videosize is not None and self.is_download:
                # 直播时无需限制文件大小
                ydl_opts['format'] += f"[filesize<{self.youtube_max_videosize}]"
            if self.youtube_max_resolution is not None:
                ydl_opts['format'] += f"[height<={self.youtube_max_resolution}]"
            ydl_opts['format'] += "+bestaudio"
            if self.youtube_prefer_acodec is not None:
                ydl_opts['format'] += f"[acodec~='^({self.youtube_prefer_acodec})']"
            # 不能由yt_dlp创建会占用文件夹
            if not os.path.exists(download_dir):
                os.makedirs(download_dir)
            with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                if not self.is_download:
                    # 直播模式不过滤但是能写入过滤
                    ydl.archive = set()
                ydl.download([self.download_url])
            # 下载成功的情况下移动到运行目录
            for file in os.listdir(download_dir):
                shutil.move(f'{download_dir}/{file}', '.')
        except DownloadError as e:
            if 'Requested format is not available' in e.msg:
                logger.error(f"{Youtube.__name__}: {self.url}: 无法获取到流，请检查vcodec,acodec,height,filesize设置")
            elif 'ffmpeg is not installed' in e.msg:
                logger.error(f"{Youtube.__name__}: {self.url}: ffmpeg未安装，无法下载")
            else:
                logger.error(f"{Youtube.__name__}: {self.url}: {e.msg}")
            return False
        finally:
            # 清理意外退出可能产生的多余文件
            try:
                del ydl
                shutil.rmtree(download_dir)
            except:
                logger.error(f"{Youtube.__name__}: {self.url}: 清理残留文件失败，请手动删除{download_dir}")
        return True


class KVFileStore:
    def __init__(self, file_path):
        self.file_path = file_path
        self.cache = {}
        self._preload_data()

    def _ensure_file_and_folder_exists(self):
        folder_path = os.path.dirname(self.file_path)
        # 如果文件夹不存在，则创建文件夹
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        # 如果文件不存在，则创建空文件
        if not os.path.exists(self.file_path):
            open(self.file_path, "w").close()

    def _preload_data(self):
        self._ensure_file_and_folder_exists()
        with open(self.file_path, "r", encoding="utf-8") as f:
            for line in f:
                k, v = line.strip().split("=")
                self.cache[k] = v

    def add(self, key, value):
        with open(self.file_path, "a", encoding="utf-8") as f:
            f.write(f"{key}={value}\n")
        # 更新缓存
        self.cache[key] = value

    def query(self, key, default=None):
        if key in self.cache:
            return self.cache[key]
        return default



import biliup.common.util
from ..common import tools
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from ..plugins import logger


@Plugin.download(regexp=r'(?:https?://)?www\.bigo\.tv')
class Bigo(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)

    async def acheck_stream(self, is_check=False):
        try:
            room_id = self.url.split('/')[-1].split('?')[0]
        except:
            logger.warning(f"{Bigo.__name__}: {self.url}: 直播间地址错误")
            return False
        try:
            room_info = (await biliup.common.util.client.post(f'https://ta.bigo.tv/official_website/studio/getInternalStudioInfo', timeout=10,
                                                             headers={**self.fake_headers, 'Accept': 'application/json'},
                                                             data={"siteId": room_id})).json()
            if room_info['code'] != 0:
                raise
        except:
            logger.warning(f"{Bigo.__name__}: {self.url}: 获取错误，本次跳过")
            return False

        try:
            if room_info['data']['alive'] is None:
                logger.warning(f"{Bigo.__name__}: {self.url}: 直播间不存在")
                return False
            if room_info['data']['alive'] != 1 or not room_info['data']['hls_src']:
                logger.debug(f"{Bigo.__name__}: {self.url}: 直播间未开播")
                return False

            self.raw_stream_url = room_info['data']['hls_src']
            self.room_title = room_info['data']['roomTopic']
        except:
            logger.warning(f"{Bigo.__name__}: {self.url}: 解析错误")
            return False

        return True



import logging
import re

from .engine.decorators import Plugin
from .plugins import general

logger = logging.getLogger('biliup')


def download(fname, url, **kwargs):
    pg = None
    for plugin in Plugin.download_plugins:
        if re.match(plugin.VALID_URL_BASE, url):
            pg = plugin(fname, url)
            for k in pg.__dict__:
                if kwargs.get(k):
                    pg.__dict__[k] = kwargs.get(k)
            break
    if not pg:
        pg = general.__plugin__(fname, url)
        logger.warning(f'Not found plugin for {fname} -> {url} This may cause problems')
    return pg.start()


def biliup_download(name, url, kwargs: dict):
    kwargs.pop('url')
    suffix = kwargs.get('format')
    if suffix:
        kwargs['suffix'] = suffix
    return download(name, url, **kwargs)


import time
import json
import re

from biliup.common.util import client
from biliup.config import config
from . import match1, logger
from biliup.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase


OFFICIAL_API = "https://api.live.bilibili.com"

@Plugin.download(regexp=r'(?:https?://)?(b23\.tv|live\.bilibili\.com)')
class Bililive(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.live_time = 0
        self.bilibili_danmaku = config.get('bilibili_danmaku', False)
        self.fake_headers['referer'] = url
        if config.get('user', {}).get('bili_cookie'):
            self.fake_headers['cookie'] = config.get('user', {}).get('bili_cookie')
        if config.get('user', {}).get('bili_cookie_file'):
            cookie_file_name = config.get('user', {}).get('bili_cookie_file')
            try:
                with open(cookie_file_name, encoding='utf-8') as stream:
                    cookies = json.load(stream)["cookie_info"]["cookies"]
                    cookies_str = ''
                    for i in cookies:
                        cookies_str += f"{i['name']}={i['value']};"
                    self.fake_headers['cookie'] = cookies_str
            except Exception:
                logger.exception("load_cookies error")
        # else:
        #    logger.warning("No cookie provided. The original quality may not be available.")

    async def acheck_stream(self, is_check=False):

        client.headers.update(self.fake_headers)

        if "b23.tv" in self.url:
            try:
                url = (await client.get(self.url, follow_redirects=False)).next_request.url
                if "live.bilibili" in str(url):
                    self.url = url
                else:
                    raise
            except:
                logger.error(f"{self.plugin_msg}: 不支持的链接")
                return False

        room_id = match1(self.url, r'bilibili.com/(\d+)')
        qualityNumber = int(config.get('bili_qn', 10000))

        # 获取直播状态与房间标题
        info_by_room_url = f"{OFFICIAL_API}/xlive/web-room/v1/index/getInfoByRoom?room_id={room_id}"
        try:
            room_info = (await client.get(info_by_room_url)).json()
        except:
            logger.exception(f"{self.plugin_msg}: ")
            return False
        if room_info['code'] != 0:
            logger.error(f"{self.plugin_msg}: {room_info}")
            return False
        if room_info['data']['room_info']['live_status'] != 1:
            logger.debug(f"{self.plugin_msg}: 未开播")
            self.raw_stream_url = None
            return False
        self.live_cover_url = room_info['data']['room_info']['cover']
        live_start_time = room_info['data']['room_info']['live_start_time']
        # 允许分段时更新标题
        self.room_title = room_info['data']['room_info']['title']
        if live_start_time > self.live_time:
            self.live_time = live_start_time
            is_new_live = True
        else:
            is_new_live = False
        if is_check:
            _res = await client.get('https://api.bilibili.com/x/web-interface/nav')
            try:
                user_data = json.loads(_res.text).get('data')
                if user_data.get('isLogin'):
                    logger.info(f"用户名：{user_data['uname']}, mid：{user_data['mid']}, isLogin：{user_data['isLogin']}")
                else:
                    logger.warning(f"{self.plugin_msg}: 未登录，或将只能录制到最低画质。")
            except:
                logger.exception(f"{self.plugin_msg}: 登录态校验失败 {_res.text}")
            return True

        protocol = config.get('bili_protocol', 'stream')
        perf_cdn = config.get('bili_perfCDN')
        cdn_fallback = config.get('bili_cdn_fallback', False)
        force_source = config.get('bili_force_source', False)
        main_api = config.get('bili_liveapi', OFFICIAL_API).rstrip('/')
        fallback_api = config.get('bili_fallback_api', OFFICIAL_API).rstrip('/')
        cn01_sids = config.get('bili_replace_cn01', [])
        if isinstance(cn01_sids, str):
            cn01_sids = cn01_sids.split(',')
        normalize_cn204 = config.get('bili_normalize_cn204', False)

        params = {
            'room_id': room_id,
            'protocol': '0,1',  # 0: http_stream, 1: http_hls
            'format': '0,1,2',  # 0: flv, 1: ts, 2: fmp4
            'codec': '0',  # 0: avc, 1: hevc, 2: av1
            'qn': qualityNumber,
            'platform': 'html5',  # web, html5, android, ios
            # 'ptype': '8',
            'dolby': '5',
            # 'panorama': '1' # 全景(不支持 html5)
        }

        if self.raw_stream_url is not None \
                and qualityNumber >= 10000 \
                and not is_new_live:
            # 同一个 streamName 即可复用，除非被超管切断
            # 前面拿不到 streamName，目前使用开播时间判断
            url = await self.acheck_url_healthy(self.raw_stream_url)
            if url is not None:
                logger.debug(f"{self.plugin_msg}: 复用 {url}")
                return True
            else:
                self.raw_stream_url = None

        try:
            play_info = await self._get_play_info(main_api, params)
            if not play_info or check_areablock(play_info):
                logger.debug(f"{self.plugin_msg}: {main_api} 返回 {play_info}")
                play_info = await self._get_play_info(fallback_api, params)
                if not play_info or check_areablock(play_info):
                    logger.debug(f"{self.plugin_msg}: {fallback_api} 返回 {play_info}")
                    return False
        except Exception:
            logger.exception(f"{self.plugin_msg}: ")
            return False
        if play_info['code'] != 0:
            logger.error(f"{self.plugin_msg}: {play_info}")
            return False

        streams = play_info['data']['playurl_info']['playurl']['stream']
        stream = streams[1] if protocol.startswith('hls') and len(streams) > 1 else streams[0]
        stream_format = stream['format'][0]
        if protocol == "hls_fmp4":
            if stream_format['format_name'] != 'fmp4':
                if len(stream['format']) > 1:
                    stream_format = stream['format'][1]
                elif int(time.time()) - live_start_time <= 60:
                    logger.warning(f"{self.plugin_msg}: 暂时未提供 hls_fmp4 流，等待下一次检测")
                    return False
                else:
                    # hls_ts 大抵是无了，只能回退 Flv
                    stream_format = streams[0]['format'][0]
                    logger.info(f"{self.plugin_msg}: 已切换为 stream 流")
        stream_info = stream_format['codec'][0]
        # 防止 hls_fmp4 不转码原画
        if qualityNumber == 10000 \
            and qualityNumber not in stream_info['accept_qn'] \
            and stream['protocol_name'] != streams[0]['protocol_name']:
            stream_info = streams[0]['format'][0]['codec'][0]
            logger.warning(
                f"{self.plugin_msg}: 当前 protocol-{protocol} 未提供原画，尝试回退到 stream 流"
            )

        stream_url = {
            'base_url': stream_info['base_url'],
        }
        if perf_cdn is not None:
            perf_cdn_list = perf_cdn.split(',')
            for url_info in stream_info['url_info']:
                if 'host' in stream_url:
                    break
                for cdn in perf_cdn_list:
                    if cdn in url_info['extra']:
                        stream_url['host'] = url_info['host']
                        stream_url['extra'] = url_info['extra']
                        logger.debug(f"Found {stream_url['host']}")
                        break
        if len(stream_url) < 3:
            stream_url['host'] = stream_info['url_info'][-1]['host']
            stream_url['extra'] = stream_info['url_info'][-1]['extra']

        # 移除 streamName 内画质标签
        if force_source:
            streamname_regexp = r"(live_\d+_\w+_\w+_?\w+?)"  # 匹配 streamName
            streamName = match1(stream_url['base_url'], streamname_regexp)
            if streamName is not None and qualityNumber >= 10000:
                _base_url = stream_url['base_url'].replace(f"_{streamName.split('_')[-1]}", '')
                if (await self.acheck_url_healthy(f"{stream_url['host']}{_base_url}{stream_url['extra']}")) is not None:
                    stream_url['base_url'] = _base_url
                else:
                    logger.debug(f"{self.plugin_msg}: force_source {_base_url}")

        if cn01_sids:
            if "cn-gotcha01" in stream_url['extra']:
                for sid in cn01_sids:
                    _host = f"https://{sid}.bilivideo.com"
                    url = f"{_host}{stream_url['base_url']}{stream_url['extra']}"
                    if (await self.acheck_url_healthy(url)) is not None:
                        stream_url['host'] = _host
                        break
                    else:
                        logger.debug(f"{self.plugin_msg}: {sid} is not available")

        self.raw_stream_url = f"{stream_url['host']}{stream_url['base_url']}{stream_url['extra']}"

        if normalize_cn204:
            self.raw_stream_url = re.sub(r"(?<=cn-gotcha204)-[1-4]", "", self.raw_stream_url, 1)

        if cdn_fallback:
            _url = await self.acheck_url_healthy(self.raw_stream_url)
            if _url is None:
                i = len(stream_info['url_info'])
                while i:
                    i -= 1
                    try:
                        self.raw_stream_url = "{}{}{}".format(stream_info['url_info'][i]['host'],
                                                              stream_url['base_url'],
                                                              stream_info['url_info'][i]['extra'])
                        _url = await self.acheck_url_healthy(self.raw_stream_url)
                        if _url is not None:
                            self.raw_stream_url = _url
                            break
                    except:
                        logger.exception("Uncaught exception:")
                        continue
                    finally:
                        logger.debug(f"{i} - {self.raw_stream_url}")
                else:
                    logger.debug(play_info)
                    self.raw_stream_url = None
                    return False
            else:
                self.raw_stream_url = _url

        return True

    def danmaku_init(self):
        if self.bilibili_danmaku:
            self.danmaku = DanmakuClient(self.url, self.gen_download_filename())


    async def _get_play_info(self, api, params) -> dict:
        api = (lambda a: a if a.startswith(('http://', 'https://')) else 'http://' + a)(api)
        full_url = f"{api}/xlive/web-room/v2/index/getRoomPlayInfo"
        try:
            _info = await client.get(full_url, params=params)
            return json.loads(_info.text)
        except:
            logger.exception(f"{params['room_id']} <- {api} 返回内容错误: {_info.text}")
        return {}


# Copy from room-player.js
def check_areablock(data):
    '''
    :return: True if area block
    '''
    if not data['data']['playurl_info']['playurl']:
        logger.error('Sorry, bilibili is currently not available in your country according to copyright restrictions.')
        logger.error('非常抱歉，根据版权方要求，您所在的地区无法观看本直播')
        return True
    return False
